#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow Adaptor Classes."""

import copy
import math
import os
import re
from collections import OrderedDict, UserDict
from copy import deepcopy
from typing import Callable, Dict

import numpy as np
import tensorflow as tf
import yaml
from pkg_resources import parse_version

from neural_compressor.common import logger
from neural_compressor.tensorflow.quantization.config import StaticQuantConfig
from neural_compressor.tensorflow.utils import (
    SPR_BASE_VERSIONS,
    UNIFY_OP_TYPE_MAPPING,
    BaseDataLoader,
    BaseModel,
    CpuInfo,
    Statistics,
    deep_get,
    dump_elapsed_time,
    singleton,
    version1_eq_version2,
    version1_gte_version2,
    version1_lt_version2,
)

spr_base_verions = SPR_BASE_VERSIONS


class TensorFlowAdaptor:
    """Adaptor Layer for stock tensorflow and spr-base."""

    unify_op_type_mapping = UNIFY_OP_TYPE_MAPPING

    def __init__(self, framework_specific_info):
        """Initialization.

        Args:
            framework_specific_info: framework specific info passed from strategy.
        """
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        self.quantize_config = {"op_wise_config": {}}
        self.framework_specific_info = framework_specific_info
        self.approach = deep_get(self.framework_specific_info, "approach", False)
        self.device = self.framework_specific_info["device"]
        self.work_dir = os.path.abspath(self.framework_specific_info["workspace_path"])
        self.recipes = deep_get(self.framework_specific_info, "recipes", {})
        self.performance_only = deep_get(self.framework_specific_info, "performance_only", False)
        self.use_bf16 = deep_get(self.framework_specific_info, "use_bf16", False)
        self.backend = self.framework_specific_info["backend"]
        self.format = self.framework_specific_info["format"]
        os.makedirs(self.work_dir, exist_ok=True)

        self.model = None
        self.pre_optimized_model = None
        self.pre_optimizer_handle = None

        self.bf16_ops = []
        self.fp32_ops = []
        self.smooth_quant_mul_ops = []
        self.dump_times = 0  # for tensorboard

        cfg_yaml_name = "{}.yaml".format(self.__class__.__name__[: -len("Adaptor")].lower())
        self.itex_mode = self.backend == "itex" or cfg_yaml_name == "tensorflow_itex.yaml"

        if self.itex_mode:  # pragma: no cover
            self._check_itex()

        self.query_handler = TensorflowQuery(
            local_config_file=os.path.join(os.path.dirname(__file__), cfg_yaml_name),
            performance_only=self.performance_only,
            itex_mode=self.itex_mode,
        )

        self.new_api = tf.version.VERSION in spr_base_verions
        self.qdq_enabled = self.itex_mode or self.format == "QDQ" or self.new_api
        self.op_wise_sequences = self.query_handler.get_eightbit_patterns(self.qdq_enabled)

        self.fp32_results = []
        self.fp32_preds_as_label = False
        # self.benchmark = GLOBAL_STATE.STATE == MODE.BENCHMARK
        self.benchmark = False
        self.callbacks = []

        self.optype_statistics = None

        self._last_dequantize_ops = None

    def _check_itex(self):  # pragma: no cover
        try:
            import intel_extension_for_tensorflow
        except:
            raise ImportError(
                "The Intel® Extension for TensorFlow is not installed. "
                "Please install it to run models on ITEX backend"
            )

    def _tuning_cfg_to_fw(self, tuning_cfg):
        """Parse the neural_compressor wrapped configuration to Tensorflow.

        Args:
            tuning_cfg (dict): configuration for quantization.
        """
        self.quantize_config["calib_iteration"] = tuning_cfg["calib_iteration"]
        self.quantize_config["device"] = self.device
        self.quantize_config["advance"] = deep_get(tuning_cfg, "advance")
        fp32_ops = []
        bf16_ops = []
        dispatched_op_names = [j[0] for j in tuning_cfg["op"]]

        invalid_op_names = [i for i in self.quantize_config["op_wise_config"] if i not in dispatched_op_names]

        for op_name in invalid_op_names:  # pragma: no cover
            self.quantize_config["op_wise_config"].pop(op_name)

        for each_op_info in tuning_cfg["op"]:
            op_name = each_op_info[0]

            if tuning_cfg["op"][each_op_info]["activation"]["dtype"] in ["fp32", "bf16"]:
                if op_name in self.quantize_config["op_wise_config"]:
                    self.quantize_config["op_wise_config"].pop(op_name)
                if tuning_cfg["op"][each_op_info]["activation"]["dtype"] == "fp32":
                    fp32_ops.append(op_name)
                if tuning_cfg["op"][each_op_info]["activation"]["dtype"] == "bf16":  # pragma: no cover
                    bf16_ops.append(op_name)
                continue

            is_perchannel = False
            bit = None
            if "weight" in tuning_cfg["op"][each_op_info]:
                is_perchannel = tuning_cfg["op"][each_op_info]["weight"]["granularity"] == "per_channel"
                # bit = tuning_cfg['op'][each_op_info]['weight']['bit']
            weight_bit = bit if bit else 7.0

            algorithm = tuning_cfg["op"][each_op_info]["activation"]["algorithm"]

            is_asymmetric = False
            if "activation" in tuning_cfg["op"][each_op_info]:
                is_asymmetric = tuning_cfg["op"][each_op_info]["activation"]["scheme"] == "asym"
            self.quantize_config["op_wise_config"][op_name] = (is_perchannel, algorithm, is_asymmetric, weight_bit)

        self.fp32_ops = fp32_ops
        self.bf16_ops = bf16_ops

    @dump_elapsed_time("Pass quantize model")
    def quantize(
        self,
        quant_config: StaticQuantConfig,
        model: BaseModel,
        calib_dataloader: Callable = None,
        calib_iteration: int = 100,
        calib_func: Callable = None,
    ):
        """Execute the quantize process on the specified model.

        Args:
            quant_config: a quantization configuration.
            model: the fp32 model to be quantized.
            calib_dataloader: a data loader for calibration.
            calib_iteration: the iteration of calibration.
            calib_func: the function used for calibration, should be a substitution for calib_dataloader
            when the built-in calibration function of INC does not work for model inference.

        Returns:
            converted_model: the quantized INC model wrapper.
        """
        assert (
            self.approach != "post_training_dynamic_quant"
        ), "Dynamic Quantization is not supported on TensorFlow framework now!"

        assert (
            self.approach != "quant_aware_training"
        ), "Quantize Aware Training is not supported on TensorFlow framework now!"

        self.calib_sampling_size = calib_dataloader.batch_size * calib_iteration if calib_dataloader else 100
        tune_cfg = self.parse_quant_config(quant_config, model, calib_iteration)
        self._tuning_cfg_to_fw(tune_cfg)
        self.bf16_ops.extend(self.smooth_quant_mul_ops)
        logger.debug("Dump quantization configurations:")
        logger.debug(self.quantize_config)
        from neural_compressor.tensorflow.quantization.utils.graph_converter import GraphConverter

        calib_sampling_size = tune_cfg.get("calib_sampling_size", 1)
        if isinstance(calib_dataloader, BaseDataLoader):
            batch_size = calib_dataloader.batch_size
            try:
                for i in range(batch_size):
                    if calib_sampling_size % (batch_size - i) == 0:
                        calib_batch_size = batch_size - i
                        if i != 0:  # pragma: no cover
                            logger.warning(
                                "Reset `calibration.dataloader.batch_size` field "
                                "to {}".format(calib_batch_size) + " to make sure the sampling_size is "
                                "divisible exactly by batch size"
                            )
                        break
                tmp_iterations = int(math.ceil(calib_sampling_size / calib_batch_size))
                calib_dataloader.batch(calib_batch_size)
                self.quantize_config["calib_iteration"] = tmp_iterations
                converted_model = GraphConverter(
                    model,
                    qt_config=self.quantize_config,
                    recipes=self.recipes,
                    int8_sequences=self.op_wise_sequences,
                    fp32_ops=self.fp32_ops,
                    bf16_ops=self.bf16_ops,
                    data_loader=calib_dataloader,
                    calib_func=calib_func,
                    qdq_enabled=self.qdq_enabled,
                    new_api=self.new_api,
                    performance_only=self.performance_only,
                    use_bf16=self.use_bf16,
                ).convert()
            except Exception:  # pragma: no cover
                from neural_compressor.tensorflow.quantization.utils.utility import get_model_input_shape

                batch_size = get_model_input_shape(model)
                logger.warning(
                    "Fail to forward with batch size={}, set to {} now.".format(calib_dataloader.batch_size, batch_size)
                )
                calib_dataloader.batch(batch_size)
                self.quantize_config["calib_iteration"] = calib_sampling_size
                converted_model = GraphConverter(
                    model,
                    qt_config=self.quantize_config,
                    recipes=self.recipes,
                    int8_sequences=self.op_wise_sequences,
                    fp32_ops=self.fp32_ops,
                    bf16_ops=self.bf16_ops,
                    data_loader=calib_dataloader,
                    calib_func=calib_func,
                    qdq_enabled=self.qdq_enabled,
                    new_api=self.new_api,
                    performance_only=self.performance_only,
                    use_bf16=self.use_bf16,
                ).convert()
        else:  # pragma: no cover
            if hasattr(calib_dataloader, "batch_size") and calib_sampling_size % calib_dataloader.batch_size != 0:
                iter = self.quantize_config["calib_iteration"]
                logger.warning(
                    "Please note that calibration sampling size {} "
                    "isn't divisible exactly by batch size {}. "
                    "So the real sampling size is {}.".format(
                        calib_sampling_size, calib_dataloader.batch_size, calib_dataloader.batch_size * iter
                    )
                )
            converted_model = GraphConverter(
                model,
                qt_config=self.quantize_config,
                recipes=self.recipes,
                int8_sequences=self.op_wise_sequences,
                fp32_ops=self.fp32_ops,
                bf16_ops=self.bf16_ops,
                data_loader=calib_dataloader,
                calib_func=calib_func,
                qdq_enabled=self.qdq_enabled,
                new_api=self.new_api,
                performance_only=self.performance_only,
                use_bf16=self.use_bf16,
            ).convert()
        # just save framework_specific_info feature for recover
        converted_model.q_config.update({"framework_specific_info": self.framework_specific_info})

        self._dump_model_op_stats(converted_model.graph_def)

        return converted_model

    def _dump_model_op_stats(self, model_graphdef):
        """Dump the whole model's OPs statistics information for analysis."""
        fp32_op_list_uint8 = copy.deepcopy(self.query_handler.get_op_types_by_precision(precision="uint8"))
        fp32_op_list_int8 = copy.deepcopy(self.query_handler.get_op_types_by_precision(precision="int8"))
        fp32_op_list = list(set(fp32_op_list_uint8).union(set(fp32_op_list_int8)))

        int8_op_prefix_list = [
            "QuantizedConv2D",
            "_FusedQuantizedConv3D",
            "QuantizedDepthwise",
            "QuantizedMaxPool",
            "QuantizedAvgPool",
            "QuantizedConcatV2",
            "QuantizedMatMul",
            "_QuantizedFusedBatchNorm",
            "_QuantizedMatMul",
            "_QuantizedBatchMatMul",
            "_QuantizedFusedInstanceNorm",
            "_FusedQuantizedDeconv2D",
            "_FusedQuantizedDeconv3D",
        ]
        from tensorflow.python.framework import dtypes

        res = {}
        for op_type in fp32_op_list:
            res[op_type] = {"INT8": 0, "BF16": 0, "FP32": 0}
        res["QuantizeV2"] = {"INT8": 0, "BF16": 0, "FP32": 0}
        res["Dequantize"] = {"INT8": 0, "BF16": 0, "FP32": 0}
        res["Cast"] = {"INT8": 0, "BF16": 0, "FP32": 0}
        fp32_op_list.extend(["QuantizeV2", "Dequantize", "Cast"])
        for i in model_graphdef.node:
            if i.op == "Const":
                continue
            possible_int8_res = [name for name in int8_op_prefix_list if i.op.find(name) != -1]

            if any(possible_int8_res):  # pragma: no cover
                origin_op_type = possible_int8_res[0].split("Quantized")[-1]
                if origin_op_type == "FusedBatchNorm":
                    origin_op_type = "FusedBatchNormV3"
                if origin_op_type == "FusedInstanceNorm":
                    origin_op_type = "_MklFusedInstanceNorm"
                if origin_op_type == "Depthwise":
                    origin_op_type = "DepthwiseConv2dNative"
                if origin_op_type == "BatchMatMul":
                    origin_op_type = "BatchMatMulV2"
                if origin_op_type == "FusedBatchMatMulV2":
                    origin_op_type = "_MklFusedBatchMatMulV2"
                if origin_op_type == "Deconv2D":
                    origin_op_type = "Conv2DBackpropInput"
                if origin_op_type == "Deconv3D":
                    origin_op_type = "Conv3DBackpropInputV2"
                res[origin_op_type]["INT8"] += 1

            if i.op in fp32_op_list:
                if "T" not in i.attr and i.op != "Cast":  # pragma: no cover
                    continue
                if i.op == "Cast":
                    if i.attr["DstT"].type == dtypes.bfloat16:
                        res[i.op]["BF16"] += 1
                    elif i.attr["DstT"].type == dtypes.float32:
                        res[i.op]["FP32"] += 1
                elif i.attr["T"].type == dtypes.bfloat16:
                    res[i.op]["BF16"] += 1
                elif i.attr["T"].type in (dtypes.quint8, dtypes.qint8):
                    res[i.op]["INT8"] += 1
                else:
                    res[i.op]["FP32"] += 1

        field_names = ["Op Type", "Total", "INT8", "BF16", "FP32"]
        output_data = [
            [op_type, sum(res[op_type].values()), res[op_type]["INT8"], res[op_type]["BF16"], res[op_type]["FP32"]]
            for op_type in fp32_op_list
        ]

        Statistics(output_data, header="Mixed Precision Statistics", field_names=field_names).print_stat()
        self.optype_statistics = field_names, output_data

    def _query_bf16_ops(self, matched_nodes):
        """Collect the bf16 OPs configuration for quantization."""
        self.bf16_op_details = OrderedDict()

        valid_precision = self.query_handler.get_mixed_precision_combination()
        if ("bf16" in valid_precision and CpuInfo().bf16) or os.getenv("FORCE_BF16") == "1":
            for details in matched_nodes:
                node_op = details[-1][0]
                node_name = details[0]

                self.bf16_op_details[(node_name, node_op)] = [
                    {"weight": {"dtype": ["bf16"]}, "activation": {"dtype": ["bf16"]}},
                    {"weight": {"dtype": "fp32"}, "activation": {"dtype": "fp32"}},
                ]

    def _query_quantizable_ops(self, matched_nodes):
        """Collect the op-wise configuration for quantization.

        Returns:
            OrderDict: op-wise configuration.
        """
        bf16_common_config = {"weight": {"dtype": "bf16"}, "activation": {"dtype": "bf16"}}
        fp32_common_config = {"weight": {"dtype": "fp32"}, "activation": {"dtype": "fp32"}}
        uint8_type = self.query_handler.get_op_types_by_precision(precision="uint8")
        int8_type = self.query_handler.get_op_types_by_precision(precision="int8")
        bf16_type = self.query_handler.get_op_types_by_precision(precision="bf16")
        tf_quantizable_op_type = list(set(uint8_type).union(set(int8_type)))

        valid_precision = self.query_handler.get_mixed_precision_combination()
        op_capability = self.query_handler.get_quantization_capability()
        conv_config = copy.deepcopy(op_capability["Conv2D"])
        conv3d_config = copy.deepcopy(op_capability["Conv3D"]) if "Conv3D" in op_capability else None
        matmul_config = copy.deepcopy(op_capability["MatMul"])
        other_config = copy.deepcopy(op_capability["default"])

        self.quantizable_op_details = OrderedDict()
        self.recipes_ops = {}

        self._init_op_stat = {i: [] for i in tf_quantizable_op_type}

        exclude_first_quantizable_op = (
            True
            if "first_conv_or_matmul_quantization" in self.recipes
            and not self.recipes["first_conv_or_matmul_quantization"]
            else False
        )
        for details in matched_nodes:
            node_op = details[-1][0]
            node_name = details[0]
            patterns = details[-1]
            pat_length = len(patterns)
            pattern_info = {
                "sequence": [[",".join(patterns[: pat_length - i]) for i in range(pat_length)][0]],
                "precision": ["int8"],
            }
            first_conv_or_matmul_node = []
            if (
                node_op in tf_quantizable_op_type
                and node_name not in self.exclude_node_names
                and (node_name, self.unify_op_type_mapping[node_op]) not in self.quantizable_op_details
            ):
                if (
                    self.unify_op_type_mapping[node_op].find("conv2d") != -1
                    or self.unify_op_type_mapping[node_op].find("matmul") != -1
                ) and len(first_conv_or_matmul_node) == 0:
                    first_conv_or_matmul_node.append((node_name, self.unify_op_type_mapping[node_op]))
                    self.recipes_ops["first_conv_or_matmul_quantization"] = first_conv_or_matmul_node
                if exclude_first_quantizable_op and (  # pragma: no cover
                    self.unify_op_type_mapping[node_op].find("conv2d") != -1
                    or self.unify_op_type_mapping[node_op].find("matmul") != -1
                ):
                    exclude_first_quantizable_op = False
                    self.exclude_node_names.append(node_name)
                    continue
                self._init_op_stat[node_op].append(node_name)
                if self.unify_op_type_mapping[node_op].find("conv2d") != -1:
                    conv2d_int8_config = copy.deepcopy(conv_config)
                    conv2d_int8_config["pattern"] = pattern_info
                    self.quantizable_op_details[(node_name, self.unify_op_type_mapping[node_op])] = [
                        conv2d_int8_config,
                        fp32_common_config,
                    ]
                elif self.unify_op_type_mapping[node_op].find("conv3d") != -1:
                    conv3d_int8_config = copy.deepcopy(conv3d_config)
                    conv3d_int8_config["pattern"] = pattern_info
                    self.quantizable_op_details[(node_name, self.unify_op_type_mapping[node_op])] = [
                        conv3d_int8_config,
                        fp32_common_config,
                    ]
                elif self.unify_op_type_mapping[node_op].find("matmul") != -1:
                    matmul_int8_config = copy.deepcopy(matmul_config)
                    matmul_int8_config["pattern"] = pattern_info
                    # TODO enable the sym mode once the tf fixed the mkldequantize_op.cc bug.
                    # is_positive_input = self.pre_optimizer_handle.has_positive_input(node_name)
                    # matmul_scheme = 'sym' if is_positive_input else 'asym'
                    matmul_scheme = ["asym"]
                    matmul_int8_config["activation"]["scheme"] = matmul_scheme
                    self.quantizable_op_details[(node_name, self.unify_op_type_mapping[node_op])] = [
                        matmul_int8_config,
                        fp32_common_config,
                    ]
                else:
                    self.quantizable_op_details[(node_name, self.unify_op_type_mapping[node_op])] = [
                        copy.deepcopy(other_config),
                        fp32_common_config,
                    ]
                if node_op in bf16_type and (
                    ("bf16" in valid_precision and CpuInfo().bf16) or os.getenv("FORCE_BF16") == "1"
                ):
                    self.quantizable_op_details[(node_name, self.unify_op_type_mapping[node_op])].insert(
                        1, bf16_common_config
                    )

                self.quantize_config["op_wise_config"][node_name] = (False, "minmax", False)
        return self.quantizable_op_details

    def _filter_unquantizable_concat(self, matched_nodes):
        """Filter out unquantizable ConcatV2 Ops based on the positive input rule."""
        target_concat_nodes = [i[0] for i in matched_nodes if i[-1][0] == "ConcatV2"]
        from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper
        from neural_compressor.tensorflow.quantization.utils.utility import GraphAnalyzer

        g = GraphAnalyzer()
        g.graph = self.pre_optimized_model.graph_def
        graph_info = g.parse_graph()
        concat_nodes = g.query_fusion_pattern_nodes([["ConcatV2"]])
        for i in concat_nodes:
            concat_node_name = i[0]
            if concat_node_name not in target_concat_nodes:  # pragma: no cover
                continue
            input_positive_status = []
            for index in range(graph_info[concat_node_name].node.attr["N"].i):
                each_input_name = GraphRewriterHelper.node_name_from_input(
                    graph_info[concat_node_name].node.input[index]
                )
                each_input_node = graph_info[each_input_name].node
                positive_input = False
                if each_input_node.op in ("Relu", "Relu6"):
                    positive_input = True
                else:
                    positive_input = g.has_positive_input(each_input_node.name)
                input_positive_status.append(positive_input)
            if not any(input_positive_status):  # pragma: no cover
                matched_nodes.remove(i)

    def _filter_unquantizable_concat_performance_only(self, matched_nodes):
        """OOB filter out unquantizable ConcatV2 OPs by checking the control flow rule."""
        target_concat_nodes = [i[0] for i in matched_nodes if i[-1][0] == "ConcatV2"]
        from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper
        from neural_compressor.tensorflow.quantization.utils.utility import GraphAnalyzer

        g = GraphAnalyzer()
        g.graph = self.pre_optimized_model.graph_def
        graph_info = g.parse_graph()
        concat_nodes = g.query_fusion_pattern_nodes([["ConcatV2"]])
        for i in concat_nodes:
            concat_node_name = i[0]
            if concat_node_name not in target_concat_nodes:  # pragma: no cover
                continue
            input_positive_status = []
            control_flow = False
            for index in range(graph_info[concat_node_name].node.attr["N"].i):
                each_input_name = GraphRewriterHelper.node_name_from_input(
                    graph_info[concat_node_name].node.input[index]
                )
                each_input_node = graph_info[each_input_name].node
                if each_input_node.op in ("Switch"):  # pragma: no cover
                    control_flow = True
            if control_flow:  # pragma: no cover
                matched_nodes.remove(i)

    def parse_quant_config(self, quant_config, model, calib_iteration):
        """Parse the quant_config to tune_cfg.

        Args:
            quant_config: a quantization configuration.
            model: the fp32 model to be quantized.
            calib_iteration: the number of iteration for calibration.

        Returns:
            tune_cfg: a dict composed by necessary information for quantization.
        """
        capability = self._query_fw_capability(model)
        config_converter = TensorflowConfigConverter(quant_config, capability)
        tune_cfg = config_converter.parse_to_tune_cfg()
        tune_cfg["calib_sampling_size"] = self.calib_sampling_size
        tune_cfg["calib_iteration"] = calib_iteration
        tune_cfg["approach"] = "post_training_static_quant"
        tune_cfg["recipe_cfgs"] = tune_cfg.get("recipe_cfgs", {})
        tune_cfg["trial_number"] = 1

        return tune_cfg

    def _query_fw_capability(self, model):
        """Collect the model-wise and op-wise configuration for quantization.

        Args:
            model (tf.compat.v1.GraphDef): model definition.

        Returns:
            [dict]: model-wise & op-wise configuration for quantization.
        """
        if self.pre_optimized_model is None:
            from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.pre_optimize import (
                PreOptimization,
            )

            self.pre_optimizer_handle = PreOptimization(model, self.new_api, self.device)
            self.pre_optimized_model = self.pre_optimizer_handle.get_optimized_model(self.itex_mode)
            model.graph_def = self.pre_optimized_model.graph_def

        self.exclude_node_names = self.pre_optimizer_handle.get_excluded_node_names()
        patterns = self.query_handler.generate_internal_patterns()
        bf16_patterns = self.query_handler.get_bf16_patterns()
        matched_nodes = self.pre_optimizer_handle.get_matched_nodes(patterns)
        matched_bf16_nodes = self.pre_optimizer_handle.get_matched_nodes(bf16_patterns)
        original_graph_node_name = [i.name for i in model.graph_def.node]
        matched_nodes = sorted(
            matched_nodes, reverse=True, key=lambda i: (original_graph_node_name.index(i[0]), len(i[-1]))
        )

        def check_match(patterns, input_pattern):
            for i in patterns:
                if input_pattern == [i for i in i.replace("+", " ").strip().split(" ") if i]:  # pragma: no cover
                    return True
            return False

        if (self.new_api and self.performance_only) or self.itex_mode or os.getenv("TF_FORCE_CONCAT_OPTS") == "1":
            self._filter_unquantizable_concat_performance_only(matched_nodes)
        else:
            self._filter_unquantizable_concat(matched_nodes)
        copied_matched_nodes = copy.deepcopy(matched_nodes)
        for i in copied_matched_nodes:
            if i[-1][0] in self.query_handler.get_op_types()["int8"]:
                continue

            if not self.pre_optimizer_handle.has_positive_input(i[0]) and not check_match(
                self.query_handler.get_fuse_patterns()["int8"], i[-1]
            ):
                matched_nodes.remove(i)

        del copied_matched_nodes

        copied_matched_nodes = copy.deepcopy(matched_bf16_nodes)
        for i in copied_matched_nodes:
            for j in matched_nodes:
                if i[0] == j[0] and i in matched_bf16_nodes:
                    matched_bf16_nodes.remove(i)

        del copied_matched_nodes

        self._query_quantizable_ops(matched_nodes)
        self._query_bf16_ops(matched_bf16_nodes)
        capability = {"optypewise": self.get_optype_wise_ability(), "recipes_ops": self.recipes_ops}
        capability["opwise"] = copy.deepcopy(self.quantizable_op_details)
        capability["opwise"].update(self.bf16_op_details)
        logger.debug("Dump framework quantization capability:")
        logger.debug(capability)

        return capability

    def quantize_input(self, model):
        """Quantize the model to be able to take quantized input.

        Remove graph QuantizedV2 op and move its input tensor to QuantizedConv2d
        and calculate the min-max scale.

        Args:
            model (tf.compat.v1.GraphDef): The model to quantize input

        Return:
            model (tf.compat.v1.GraphDef): The quantized input model
            scale (float): The scale for dataloader to generate quantized input
        """
        scale = None
        # quantize input only support tensorflow version > 2.1.0
        if version1_lt_version2(tf.version.VERSION, "2.1.0"):  # pragma: no cover
            logger.warning("Quantize input needs tensorflow 2.1.0 and newer.")
            return model, scale

        graph_def = model.as_graph_def()
        node_name_mapping = {}
        quantize_nodes = []
        for node in graph_def.node:
            node_name_mapping[node.name] = node
            if node.op == "QuantizeV2":
                quantize_nodes.append(node)

        target_quantize_nodes = []
        for node in quantize_nodes:
            # only support Quantizev2 input op Pad and Placeholder
            if (
                node_name_mapping[node.input[0]].op == "Pad"
                and node_name_mapping[node_name_mapping[node.input[0]].input[0]].op == "Placeholder"
            ) or node_name_mapping[node.input[0]].op == "Placeholder":
                target_quantize_nodes.append(node)
        assert len(target_quantize_nodes) == 1, "only support 1 QuantizeV2 from Placeholder"
        quantize_node = target_quantize_nodes[0]

        quantize_node_input = node_name_mapping[quantize_node.input[0]]
        quantize_node_outputs = [node for node in graph_def.node if quantize_node.name in node.input]

        from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper

        if quantize_node_input.op == "Pad":
            pad_node_input = node_name_mapping[quantize_node_input.input[0]]
            assert pad_node_input.op == "Placeholder", "only support Pad between QuantizeV2 and Placeholder"
            from tensorflow.python.framework import tensor_util

            paddings_tensor = tensor_util.MakeNdarray(
                node_name_mapping[quantize_node_input.input[1]].attr["value"].tensor
            ).flatten()

            quantize_node.input[0] = quantize_node_input.input[0]
            for conv_node in quantize_node_outputs:
                assert "Conv2D" in conv_node.op, "only support QuantizeV2 to Conv2D"

                GraphRewriterHelper.set_attr_int_list(conv_node, "padding_list", paddings_tensor)
            graph_def.node.remove(quantize_node_input)

        from tensorflow.python.framework import dtypes

        GraphRewriterHelper.set_attr_dtype(node_name_mapping[quantize_node.input[0]], "dtype", dtypes.qint8)

        for conv_node in quantize_node_outputs:
            for index, conv_input in enumerate(conv_node.input):
                if conv_input == quantize_node.name:
                    conv_node.input[index] = quantize_node.input[0]
                elif conv_input == quantize_node.name + ":1":
                    conv_node.input[index] = quantize_node.input[1]
                elif conv_input == quantize_node.name + ":2":
                    conv_node.input[index] = quantize_node.input[2]

        # get the input's min-max value and calculate scale
        max_node = node_name_mapping[quantize_node.input[2]]
        min_node = node_name_mapping[quantize_node.input[1]]
        max_value = max_node.attr["value"].tensor.float_val[0]
        min_value = min_node.attr["value"].tensor.float_val[0]
        scale = 127.0 / max(abs(max_value), abs(min_value))
        # remove QuantizeV2 node
        graph_def.node.remove(quantize_node)

        graph = tf.Graph()
        with graph.as_default():
            # use name='' to avoid 'import/' to name scope
            tf.import_graph_def(graph_def, name="")
        return graph, scale

    def get_optype_wise_ability(self):
        """Get the op type wise capability by generating the union value of each op type.

        Returns:
            [string dict]: the key is op type while the value is the
                           detail configurations of activation and weight for this op type.
        """
        res = OrderedDict()
        for op in self.quantizable_op_details:
            if op[1] not in res:
                res[op[1]] = {"activation": self.quantizable_op_details[op][0]["activation"]}
                if "weight" in self.quantizable_op_details[op][0]:
                    res[op[1]]["weight"] = self.quantizable_op_details[op][0]["weight"]
        for op in self.bf16_op_details:
            if op[1] not in res:
                res[op[1]] = {"activation": {"dtype": ["bf16"]}, "weight": {"dtype": ["bf16"]}}
        return res


class Tensorflow_ITEXAdaptor(TensorFlowAdaptor):  # pragma: no cover
    """Tensorflow ITEX Adaptor Class."""

    def __init__(self, framework_specific_info):
        """Initialization.

        Args:
            framework_specific_info: framework specific information.
        """
        super().__init__(framework_specific_info)

    @dump_elapsed_time("Pass quantize model")
    def quantize(
        self,
        quant_config: StaticQuantConfig,
        model: BaseModel,
        calib_dataloader: Callable = None,
        calib_iteration: int = 100,
        calib_func: Callable = None,
    ):
        """Execute the quantize process on the specified model.

        Args:
            quant_config: a quantization configuration.
            model: the fp32 model to be quantized.
            calib_dataloader: a data loader for calibration.
            calib_iteration: the iteration of calibration.
            calib_func: the function used for calibration, should be a substitution for calib_dataloader
            when the built-in calibration function of INC does not work for model inference.

        Returns:
            converted_model: the quantized INC model wrapper.
        """
        self.calib_sampling_size = calib_dataloader.batch_size * calib_iteration
        tune_cfg = self.parse_quant_config(quant_config, model, calib_iteration)
        self._tuning_cfg_to_fw(tune_cfg)
        logger.debug("Dump quantization configurations:")
        logger.debug(self.quantize_config)
        from neural_compressor.tensorflow.quantization.utils.graph_converter import GraphConverter

        self.calib_sampling_size = tune_cfg.get("calib_sampling_size", 1)
        if isinstance(calib_dataloader, BaseDataLoader):
            batch_size = calib_dataloader.batch_size
            try:
                for i in range(batch_size):
                    if self.calib_sampling_size % (batch_size - i) == 0:
                        calib_batch_size = batch_size - i
                        if i != 0:  # pragma: no cover
                            logger.warning(
                                "Reset `calibration.dataloader.batch_size` field "
                                "to {}".format(calib_batch_size) + " to make sure the sampling_size is "
                                "divisible exactly by batch size"
                            )
                        break
                tmp_iterations = int(math.ceil(self.calib_sampling_size / calib_batch_size))
                calib_dataloader.batch(calib_batch_size)
                self.quantize_config["calib_iteration"] = tmp_iterations

                converted_model = GraphConverter(
                    model,
                    qt_config=self.quantize_config,
                    recipes=self.recipes,
                    int8_sequences=self.op_wise_sequences,
                    fp32_ops=self.fp32_ops,
                    bf16_ops=self.bf16_ops,
                    data_loader=calib_dataloader,
                    calib_func=calib_func,
                    itex_mode=self.itex_mode,
                    qdq_enabled=self.qdq_enabled,
                    new_api=self.new_api,
                    performance_only=self.performance_only,
                    use_bf16=self.use_bf16,
                ).convert()
            except Exception:  # pragma: no cover
                from neural_compressor.tensorflow.quantization.utils.utility import get_model_input_shape

                batch_size = get_model_input_shape(model)
                logger.warning(
                    "Fail to forward with batch size={}, set to {} now.".format(calib_dataloader.batch_size, batch_size)
                )
                calib_dataloader.batch(batch_size)
                self.quantize_config["calib_iteration"] = self.calib_sampling_size
                converted_model = GraphConverter(
                    model,
                    qt_config=self.quantize_config,
                    recipes=self.recipes,
                    int8_sequences=self.op_wise_sequences,
                    fp32_ops=self.fp32_ops,
                    bf16_ops=self.bf16_ops,
                    data_loader=calib_dataloader,
                    itex_mode=self.itex_mode,
                    qdq_enabled=self.qdq_enabled,
                    new_api=self.new_api,
                    performance_only=self.performance_only,
                    use_bf16=self.use_bf16,
                ).convert()
        else:  # pragma: no cover
            if hasattr(calib_dataloader, "batch_size") and self.calib_sampling_size % calib_dataloader.batch_size != 0:
                iter = self.quantize_config["calib_iteration"]
                logger.warning(
                    "Please note that calibration sampling size {} "
                    "isn't divisible exactly by batch size {}. "
                    "So the real sampling size is {}.".format(
                        self.calib_sampling_size, calib_dataloader.batch_size, calib_dataloader.batch_size * iter
                    )
                )
            converted_model = GraphConverter(
                model,
                qt_config=self.quantize_config,
                recipes=self.recipes,
                int8_sequences=self.op_wise_sequences,
                fp32_ops=self.fp32_ops,
                bf16_ops=self.bf16_ops,
                data_loader=calib_dataloader,
                calib_func=calib_func,
                itex_mode=self.itex_mode,
                qdq_enabled=self.qdq_enabled,
                new_api=self.new_api,
                performance_only=self.performance_only,
                use_bf16=self.use_bf16,
            ).convert()

        self._dump_model_op_stats(converted_model.graph_def)

        return converted_model


class TensorFlowConfig:
    """Base config class for TensorFlow."""

    def __init__(self, precisions=None):
        """Init an TF object."""
        self._precisions = precisions

    @property
    def precisions(self):
        """Get precision."""
        return self._precisions

    @precisions.setter
    def precisions(self, precisions):  # pragma: no cover
        """Set precision."""
        if not isinstance(precisions, list):
            precisions = [precisions]
        for pr in precisions:
            self.check_value("precisions", pr, str, ["int8", "uint8", "fp32", "bf16", "fp16"])
        self._precisions = precisions

    @staticmethod
    def check_value(name, src, supported_type, supported_value=[]):  # pragma: no cover
        """Check if the given object is the given supported type and in the given supported value.

        Example::
            def datatype(self, datatype):
                if self.check_value("datatype", datatype, list, ["fp32", "bf16", "uint8", "int8"]):
                    self._datatype = datatype
        """
        if isinstance(src, list) and any([not isinstance(i, supported_type) for i in src]):
            assert False, "Type of {} items should be {} but not {}".format(
                name, str(supported_type), [type(i) for i in src]
            )
        elif not isinstance(src, list) and not isinstance(src, supported_type):
            assert False, "Type of {} should be {} but not {}".format(name, str(supported_type), type(src))

        if len(supported_value) > 0:
            if isinstance(src, str) and src not in supported_value:
                assert False, "{} is not in supported {}: {}. Skip setting it.".format(src, name, str(supported_value))
            elif (
                isinstance(src, list)
                and all([isinstance(i, str) for i in src])
                and any([i not in supported_value for i in src])
            ):
                assert False, "{} is not in supported {}: {}. Skip setting it.".format(src, name, str(supported_value))

        return True


class TensorflowQuery:
    """Tensorflow Query Capability Class."""

    def __init__(self, local_config_file=None, performance_only=False, itex_mode=False, quant_mode="static"):
        """Initialization.

        Args:
            local_config_file: local configuration file name.
            performance_only: oob performance only mode.
            itex_mode: check if itex mode.
            quant_mode: quantization mode, static or dynamic.
        """
        self.version = tf.version.VERSION
        self.cfg = local_config_file
        self.cur_config = None
        self.performance_only = performance_only
        self.quant_mode = quant_mode
        self.itex_mode = itex_mode
        self._one_shot_query()

    def _get_specified_version_cfg(self, data):
        """Get the configuration for the current runtime.

        If there's no matched configuration in the input yaml, we'll
        use the configuration of the nearest framework version field of yaml.

        Args:
            data (Yaml content): input yaml file.

        Returns:
            [dictionary]: the content for specific version.
        """
        from functools import cmp_to_key

        config = None

        def _compare(version1, version2):
            if parse_version(version1) == parse_version(version2):  # pragma: no cover
                return 0
            elif parse_version(version1) < parse_version(version2):
                return -1
            else:
                return 1

        fallback_list = []
        for sub_data in data:
            if "default" in sub_data["version"]["name"]:
                assert config is None, "Only one default config " "is allowed in framework yaml file."
                config = sub_data

            if self.version in sub_data["version"]["name"]:
                return sub_data
            else:
                if sub_data["version"]["name"] == [
                    "2.11.0202242",
                    "2.11.0202250",
                    "2.11.0202317",
                    "2.11.0202323",
                    "2.14.0202335",
                    "2.14.dev202335",
                    "2.15.0202341",
                ]:
                    continue
                sorted_list = copy.deepcopy(sub_data["version"]["name"])
                sorted_list.remove("default") if "default" in sorted_list else None
                if isinstance(sorted_list, list):
                    # TensorFlow 1.15.0-up1/up2/up3 release versions are abnoraml release naming
                    # convention. Replacing them with dot for version comparison.
                    sorted_list = [i.replace("-up", ".") for i in sorted_list]
                    sorted_list = sorted(sorted_list, key=cmp_to_key(_compare), reverse=True)
                else:  # pragma: no cover
                    assert isinstance(sorted_list, str)
                    sorted_list = list(sorted_list.replace("-up", ".").split())
                for i in sorted_list:
                    if parse_version(self.version) >= parse_version(i):
                        fallback_list.append([i, sub_data])
                        break

        assert config is not None, "The default config in framework yaml must exist."
        nearest_version = str(0)
        for fallback in fallback_list:
            if parse_version(fallback[0]) > parse_version(nearest_version):
                nearest_version = fallback[0]
                config = fallback[1]

        return config

    def _one_shot_query(self):
        """One short query for some patterns."""
        # pylint: disable=E1136
        with open(self.cfg) as f:
            content = yaml.safe_load(f)
            try:
                self.cur_config = self._get_specified_version_cfg(content)
                if not self.performance_only:
                    remove_int8_ops = [
                        "FusedBatchNorm",
                        "FusedBatchNormV2",
                        "FusedBatchNormV3",
                        "_MklFusedInstanceNorm",
                    ]
                    for op_type in remove_int8_ops:
                        while op_type in self.cur_config["int8"][self.quant_mode].keys():
                            self.cur_config["int8"][self.quant_mode].pop(op_type, None)

            except Exception as e:
                logger.info("Fail to parse {} due to {}.".format(self.cfg, str(e)))
                self.cur_config = None
                raise ValueError(
                    "Please check if the format of {} follows Neural Compressor yaml schema.".format(self.cfg)
                )
        self._update_cfg_with_usr_definition()

    def _update_cfg_with_usr_definition(self):
        """Add user defined precision configuration."""
        tensorflow_config = TensorFlowConfig()
        if tensorflow_config.precisions is not None:  # pragma: no cover
            self.cur_config["precisions"]["names"] = ",".join(tensorflow_config.precisions)

    def get_version(self):
        """Get the current backend version information.

        Returns:
            [string]: version string.
        """
        return self.cur_config["version"]["name"]

    def get_op_types(self):
        """Get the supported op types by all precisions.

        Returns:
            [dictionary list]: A list composed of dictionary which key is precision
            and value is the op types.
        """
        return {
            "int8": self.get_op_types_by_precision("int8"),
            "uint8": self.get_op_types_by_precision("uint8"),
            "bf16": self.get_op_types_by_precision("bf16"),
        }

    def get_fuse_patterns(self):
        """Get supported patterns by low precisions.

        Returns:
            [dictionary list]: A list composed of dictionary which key is precision
            and value is the supported patterns.
        """
        spr_int8_pattern_list = [
            "Conv2D + BiasAdd",
            "Conv2D + BiasAdd + Add + Relu6 + Mul + Mul",
            "Conv2D + Add + Relu6 + Mul + Mul",
            "Conv2D + BiasAdd + swish_f32",
            "Conv2D + Add + swish_f32",
            "Conv2D + AddV2 + swish_f32",
            "Conv2D + swish_f32",
            "Conv2D + BiasAdd + Relu",
            "Conv2D + Relu",
            "Conv2D + BiasAdd + Elu",
            "Conv2D + Elu",
            "Conv2D + BiasAdd + Relu6",
            "Conv2D + Relu6",
            "Conv2D + BiasAdd + LeakyRelu",
            "Conv2D + BiasAdd + Add + LeakyRelu",
            "Conv2D + BiasAdd + AddV2 + LeakyRelu",
            "Conv2D + Add + LeakyRelu",
            "Conv2D + AddV2 + LeakyRelu",
            "Conv2D + LeakyRelu",
            "Conv2D + BiasAdd + Sigmoid",
            "Conv2D + Sigmoid",
            "Conv2D + BiasAdd + LeakyRelu + AddV2",
            "Conv2D + BiasAdd + LeakyRelu + Add",
            "Conv2D + LeakyRelu + AddV2",
            "Conv2D + LeakyRelu + Add",
            "Conv2D + BiasAdd + Relu + AddV2",
            "Conv2D + BiasAdd + Relu + Add",
            "Conv2D + Relu + AddV2",
            "Conv2D + Relu + Add",
            "Conv2D + Add",
            "Conv2D + AddV2",
            "Conv2D + AddV2 + Add",
            "Conv2D + Add + Add",
            "Conv2D + BiasAdd + Add",
            "Conv3D + Add",
            "Conv3D + AddV2",
            "Conv3D + BiasAdd",
            "Conv3D + BiasAdd + Add",
            "Conv3D + BiasAdd + AddV2",
            "Conv3D + AddV2 + AddV2",
            "DepthwiseConv2dNative + BiasAdd + Add + Relu6 + Mul + Mul",
            "DepthwiseConv2dNative + Add + Relu6 + Mul + Mul",
            "DepthwiseConv2dNative + BiasAdd + swish_f32",
            "DepthwiseConv2dNative + Add + swish_f32",
            "DepthwiseConv2dNative + AddV2 + swish_f32",
            "DepthwiseConv2dNative + swish_f32",
            "DepthwiseConv2dNative + BiasAdd + LeakyRelu",
            "DepthwiseConv2dNative + LeakyRelu",
            "DepthwiseConv2dNative + BiasAdd + Relu6",
            "DepthwiseConv2dNative + Relu6",
            "DepthwiseConv2dNative + BiasAdd + Relu",
            "DepthwiseConv2dNative + Relu",
            "DepthwiseConv2dNative + Add + Relu6",
            "DepthwiseConv2dNative + BiasAdd",
            "FusedBatchNormV3 + Relu",
            "FusedBatchNormV3 + LeakyRelu",
            "_MklFusedInstanceNorm + Relu",
            "_MklFusedInstanceNorm + LeakyRelu",
            "Conv2DBackpropInput + BiasAdd",
            "Conv3DBackpropInputV2 + BiasAdd",
        ]

        spr_uint8_pattern_list = [
            "Conv2D + BiasAdd + AddN + Relu",
            "Conv2D + AddN + Relu",
            "Conv2D + BiasAdd + AddN + Relu6",
            "Conv2D + AddN + Relu6",
            "Conv2D + BiasAdd + AddV2 + Relu",
            "Conv2D + AddV2 + Relu",
            "Conv2D + BiasAdd + AddV2 + Relu6",
            "Conv2D + AddV2 + Relu6",
            "Conv2D + BiasAdd + Add + Relu",
            "Conv2D + Add + Relu",
            "Conv2D + BiasAdd + Add + Relu6",
            "Conv2D + Add + Relu6",
            "Conv2D + BiasAdd + Relu",
            "Conv2D + BiasAdd + Relu6",
            "Conv2D + Relu",
            "Conv2D + Relu6",
            "Conv2D + BiasAdd",
            "Conv2D + Add + Add + Relu",
            "DepthwiseConv2dNative + BiasAdd + Relu6",
            "DepthwiseConv2dNative + Relu6",
            "DepthwiseConv2dNative + BiasAdd + Relu",
            "DepthwiseConv2dNative + Relu",
            "DepthwiseConv2dNative + Add + Relu6",
            "DepthwiseConv2dNative + BiasAdd",
            "MatMul + BiasAdd",
            "MatMul + BiasAdd + Add",
            "MatMul + BiasAdd + AddV2",
            "MatMul + BiasAdd + Relu",
            "MatMul + BiasAdd + Relu6",
            "MatMul + BiasAdd + LeakyRelu",
            "MatMul + BiasAdd + Gelu",
            "MatMul + BiasAdd + Elu",
            "MatMul + BiasAdd + Tanh",
            "MatMul + BiasAdd + Sigmoid",
            "MatMul + Add",
            "MatMul + AddV2",
            "MatMul + Relu",
            "MatMul + Relu6",
            "MatMul + LeakyRelu",
            "MatMul + Gelu",
            "MatMul + Elu",
            "MatMul + Tanh",
            "MatMul + Sigmoid",
            "BatchMatMul + Mul",
            "BatchMatMulV2 + Mul",
            "BatchMatMul + Add",
            "BatchMatMulV2 + Add",
            "BatchMatMul + AddV2",
            "BatchMatMulV2 + AddV2",
            "BatchMatMul + Mul + Add",
            "BatchMatMulV2 + Mul + Add",
            "BatchMatMul + Mul + AddV2",
            "BatchMatMulV2 + Mul + AddV2",
            "Conv3D + AddV2 + AddV2 + Relu",
            "Conv3D + Add + Relu",
            "Conv3D + AddV2 + Relu",
            "Conv3D + Relu",
            "Conv3D + Relu6",
            "Conv3D + Add + Relu6",
            "Conv3D + AddV2 + Relu6",
            "Conv3D + Elu",
            "Conv3D + LeakyRelu",
            "Conv3D + BiasAdd + Relu",
            "Conv3D + BiasAdd + Relu6",
            "Conv3D + BiasAdd + Elu",
            "Conv3D + BiasAdd + LeakyRelu",
            "Conv3D + Add + Elu",
            "Conv3D + Add + LeakyRelu",
            "Conv2DBackpropInput + BiasAdd",
            "Conv3DBackpropInputV2 + BiasAdd",
        ]

        tf_int8_pattern_list = ["Conv2D + BiasAdd", "Conv2D + BiasAdd + Relu", "Conv2D + BiasAdd + Relu6"]
        tf_uint8_pattern_list = [
            "Conv2D + BiasAdd + AddN + Relu",
            "Conv2D + BiasAdd + AddN + Relu6",
            "Conv2D + BiasAdd + AddV2 + Relu",
            "Conv2D + BiasAdd + AddV2 + Relu6",
            "Conv2D + BiasAdd + Add + Relu",
            "Conv2D + BiasAdd + Add + Relu6",
            "Conv2D + BiasAdd + Relu",
            "Conv2D + BiasAdd + Relu6",
            "Conv2D + Add + Relu",
            "Conv2D + Add + Relu6",
            "Conv2D + Relu",
            "Conv2D + Relu6",
            "Conv2D + BiasAdd",
            "DepthwiseConv2dNative + BiasAdd + Relu6",
            "DepthwiseConv2dNative + BiasAdd + Relu",
            "DepthwiseConv2dNative + Add + Relu6",
            "DepthwiseConv2dNative + BiasAdd",
            "MatMul + BiasAdd + Relu",
            "MatMul + BiasAdd",
        ]
        tf1_15_up3_int8_pattern_list = [
            "Conv2D + BiasAdd",
            "Conv2D + BiasAdd + Relu",
            "Conv2D + BiasAdd + LeakyRelu",
            "Conv2D + BiasAdd + LeakyRelu + AddV2",
            "Conv2D + BiasAdd + Relu6",
        ]
        tf1_15_up3_uint8_pattern_list = [
            "Conv2D + BiasAdd + AddN + Relu",
            "Conv2D + BiasAdd + AddN + Relu6",
            "Conv2D + BiasAdd + AddV2 + Relu",
            "Conv2D + BiasAdd + AddV2 + Relu6",
            "Conv2D + BiasAdd + Add + Relu",
            "Conv2D + BiasAdd + Add + Relu6",
            "Conv2D + BiasAdd + Relu",
            "Conv2D + BiasAdd + Relu6",
            "Conv2D + Add + Relu",
            "Conv2D + Add + Relu6",
            "Conv2D + Relu",
            "Conv2D + Relu6",
            "Conv2D + BiasAdd",
            "DepthwiseConv2dNative + BiasAdd + Relu6",
            "DepthwiseConv2dNative + Add + Relu6",
            "DepthwiseConv2dNative + BiasAdd",
            "MatMul + BiasAdd + Relu",
            "MatMul + BiasAdd",
        ]
        old_tf_int8_pattern_list = ["MatMul + BiasAdd + Relu", "MatMul + BiasAdd"]
        old_tf_uint8_pattern_list = [
            "Conv2D + BiasAdd + AddN + Relu",
            "Conv2D + BiasAdd + AddN + Relu6",
            "Conv2D + BiasAdd + AddV2 + Relu",
            "Conv2D + BiasAdd + AddV2 + Relu6",
            "Conv2D + BiasAdd + Add + Relu",
            "Conv2D + BiasAdd + Add + Relu6",
            "Conv2D + BiasAdd + Relu",
            "Conv2D + BiasAdd + Relu6",
            "Conv2D + Add + Relu",
            "Conv2D + Add + Relu6",
            "Conv2D + Relu",
            "Conv2D + Relu6",
            "Conv2D + BiasAdd",
            "DepthwiseConv2dNative + BiasAdd + Relu6",
            "DepthwiseConv2dNative + Add + Relu6",
            "DepthwiseConv2dNative + BiasAdd",
            "MatMul + BiasAdd + Relu",
            "MatMul + BiasAdd",
        ]

        for index, pattern in enumerate(spr_int8_pattern_list):
            spr_int8_pattern_list[index] = "Dequantize + " + pattern + " + QuantizeV2"
        for index, pattern in enumerate(spr_uint8_pattern_list):
            spr_uint8_pattern_list[index] = "Dequantize + " + pattern + " + QuantizeV2"

        if not self.performance_only:
            remove_int8_ops = ["FusedBatchNorm", "FusedBatchNormV2", "FusedBatchNormV3", "_MklFusedInstanceNorm"]
            for op_type in remove_int8_ops:
                patterns = [
                    f"Dequantize + {op_type} + Relu + QuantizeV2",
                    f"Dequantize + {op_type} + LeakyRelu + QuantizeV2",
                ]
                for pattern in patterns:
                    while pattern in spr_int8_pattern_list:
                        spr_int8_pattern_list.remove(pattern)
                    while pattern in spr_uint8_pattern_list:
                        spr_uint8_pattern_list.remove(pattern)

        patterns = {}
        if tf.version.VERSION in spr_base_verions or self.itex_mode:
            patterns["int8"] = spr_int8_pattern_list
            patterns["uint8"] = spr_uint8_pattern_list
        elif version1_gte_version2(tf.version.VERSION, "2.1.0"):
            patterns["int8"] = tf_int8_pattern_list
            patterns["uint8"] = tf_uint8_pattern_list
            if self.itex_mode:  # pragma: no cover
                patterns["int8"].append("FusedBatchNormV3 + Relu")
                patterns["int8"].append("FusedBatchNormV3 + LeakyRelu")
        elif version1_eq_version2(tf.version.VERSION, "1.15.0-up3"):  # pragma: no cover
            patterns["int8"] = tf1_15_up3_int8_pattern_list
            patterns["uint8"] = tf1_15_up3_uint8_pattern_list
        else:  # pragma: no cover
            patterns["int8"] = old_tf_int8_pattern_list
            patterns["uint8"] = old_tf_uint8_pattern_list

        return patterns

    def get_quantization_capability(self):
        """Get the supported op types' quantization capability.

        Returns:
            [dictionary list]: A list composed of dictionary which key is precision
            and value is a dict that describes all op types' quantization capability.
        """
        for op_type, _ in self.cur_config["int8"][self.quant_mode].items():
            self.cur_config["int8"][self.quant_mode][op_type]["activation"]["quant_mode"] = self.quant_mode
        return self.cur_config["int8"][self.quant_mode]

    def get_op_types_by_precision(self, precision):
        """Get op types per precision.

        Args:
            precision (string): precision name

        Returns:
            [string list]: A list composed of op type.
        """
        assert precision in ("bf16", "uint8", "int8")

        if precision == "int8":
            if tf.version.VERSION in spr_base_verions or self.itex_mode:
                op_type_list = [key for key in self.cur_config["int8"][self.quant_mode].keys()]
                if not self.performance_only and not self.itex_mode:
                    remove_int8_ops = [
                        "FusedBatchNorm",
                        "FusedBatchNormV2",
                        "FusedBatchNormV3",
                        "_MklFusedInstanceNorm",
                    ]
                    for op_type in remove_int8_ops:
                        while op_type in op_type_list:
                            op_type_list.remove(op_type)
                return op_type_list
            if version1_gte_version2(tf.version.VERSION, "2.1.0") or version1_eq_version2(
                tf.version.VERSION, "1.15.0-up3"
            ):
                return ["Conv2D", "MatMul", "ConcatV2", "MaxPool", "AvgPool"]
            return ["MatMul", "ConcatV2", "MaxPool", "AvgPool"]  # pragma: no cover
        if precision == "uint8":
            if tf.version.VERSION in spr_base_verions:
                return [key for key in self.cur_config["int8"][self.quant_mode].keys() if "Norm" not in key]
            if version1_gte_version2(tf.version.VERSION, "2.1.0") or version1_eq_version2(
                tf.version.VERSION, "1.15.0-up3"
            ):
                return ["Conv2D", "MatMul", "ConcatV2", "MaxPool", "AvgPool", "DepthwiseConv2dNative"]
            return ["Conv2D", "MatMul", "ConcatV2", "MaxPool", "AvgPool"]  # pragma: no cover
        if precision == "bf16":
            if tf.version.VERSION in spr_base_verions:
                return self.cur_config[precision]
            if version1_gte_version2(tf.version.VERSION, "2.1.0") or version1_eq_version2(
                tf.version.VERSION, "1.15.0-up3"
            ):
                return self.cur_config[precision]
            return []  # pragma: no cover

    def get_mixed_precision_combination(self):
        """Get the valid mixed precisions.

        Returns:
            [string list]: valid precision list.
        """
        if version1_gte_version2(tf.version.VERSION, "2.1.0") or version1_eq_version2(tf.version.VERSION, "1.15.0-up3"):
            return ["int8", "uint8", "bf16", "fp32"]
        return ["uint8", "fp32"]

    def get_bf16_patterns(self):
        """Get BF16 pattern list.

        Returns:
            [List]: bf16 pattern list.
        """
        bf16_op_types = [i for i in self.get_op_types_by_precision("bf16")]
        res = []
        for i in bf16_op_types:
            res.append([[i]])

        return res

    def get_eightbit_patterns(self, qdq_enabled=False):
        """Get eightbit op wise sequences information.

        Returns:
            [dictionary]: key is the op type while value is the list of sequences start
                        with the op type same as key value.
        """
        quantizable_op_types = self.get_op_types_by_precision("int8") + self.get_op_types_by_precision("uint8")
        int8_patterns = [
            i.replace("+", " ").split()
            for i in list(set(self.get_fuse_patterns()["int8"] + self.get_fuse_patterns()["uint8"]))
        ]
        res = {}
        for i in quantizable_op_types:
            if qdq_enabled:
                res[i] = [["Dequantize", i, "QuantizeV2"]]
            else:
                res[i] = [[i]]

        for pattern in int8_patterns:
            if qdq_enabled:
                op_type = pattern[1]
            else:
                op_type = pattern[0]
            if op_type in res:
                res[op_type].append(pattern)

        return res

    def generate_internal_patterns(self):
        """Translate the patterns defined in the yaml to internal pattern expression."""

        def _generate_pattern(data):
            length = [len(i) for i in data]
            res = []
            for index in range(max(length)):
                if index <= min(length) - 1:
                    tmp = [i[index] for i in data]
                    if len(set(tmp)) == 1:
                        res.append([tmp[0]])
                    else:
                        res.append(tuple(set(tmp)))
                else:
                    tmp1 = [i[index] for i in data if len(i) > index]
                    res.append(tuple(set(tmp1)))

            return res

        op_level_sequences = {}

        for k, op_level_all_sequences in self.get_eightbit_patterns().items():
            op_level_sequences[k] = []
            sorted_sequences = sorted(op_level_all_sequences)
            last_len = 1
            each_combination = []
            for index, value in enumerate(sorted_sequences):
                if len(value) >= last_len:
                    last_len = len(value)
                    each_combination.append(value)
                else:
                    op_level_sequences[k].append(copy.deepcopy(each_combination))
                    each_combination.clear()
                    each_combination.append(value)
                    last_len = len(value)

                if index == len(sorted_sequences) - 1:
                    op_level_sequences[k].append(copy.deepcopy(each_combination))

        final_out = []
        for _, op_level_sequences in op_level_sequences.items():
            for similar_sequences in op_level_sequences:
                final_out.append(_generate_pattern(similar_sequences))

        return final_out


class TensorflowConfigConverter:
    """Convert `StaticQuantConfig` to the format used by static quant algo."""

    unify_op_type_mapping = UNIFY_OP_TYPE_MAPPING

    def __init__(self, quant_config: StaticQuantConfig, capability: Dict):
        """Init parser for TF static quant config.

        Args:
            quant_config: the keras static quant config.
            capability: the supported config lists for each op.
        """
        self.quant_config = quant_config
        self.capability = capability

    def update_opwise_config(self):
        """Update op-wise config.

        Args:
            quant_config: the Tensorflow static quant config.
        """
        op_wise_config = {}
        for op_name, op_config in self.quant_config.items():
            op_key_name = (op_name[0], self.unify_op_type_mapping[op_name[1]])
            if op_key_name not in self.capability["opwise"]:
                continue
            single_op_cap = self.capability["opwise"][op_key_name][0]
            single_op_config = {"activation": {}}

            single_op_config["activation"]["dtype"] = (
                op_config.act_dtype
                if op_config.act_dtype in single_op_cap["activation"]["dtype"]
                or op_config.act_dtype in ("fp32", "bf16")
                else single_op_cap["activation"]["dtype"][0]
            )

            single_op_config["activation"]["scheme"] = "sym" if op_config.act_sym else "asym"
            if single_op_config["activation"]["scheme"] not in single_op_cap["activation"]["scheme"]:
                single_op_config["activation"]["scheme"] = single_op_cap["activation"]["scheme"][0]

            single_op_config["activation"]["granularity"] = (
                op_config.act_granularity
                if op_config.act_granularity in single_op_cap["activation"]["granularity"]
                else single_op_cap["activation"]["granularity"][0]
            )

            single_op_config["activation"]["algorithm"] = (
                op_config.act_algorithm
                if op_config.act_algorithm in single_op_cap["activation"]["algorithm"]
                else single_op_cap["activation"]["algorithm"][0]
            )

            if "weight" not in single_op_cap:
                op_wise_config.update({op_key_name: single_op_config})
                continue

            single_op_config["weight"] = {}
            single_op_config["weight"]["dtype"] = (
                op_config.weight_dtype
                if op_config.weight_dtype in single_op_cap["weight"]["dtype"]
                or op_config.weight_dtype in ("fp32", "bf16")
                else single_op_cap["weight"]["dtype"][0]
            )

            single_op_config["weight"]["scheme"] = "sym" if op_config.weight_sym else "asym"
            if single_op_config["weight"]["scheme"] not in single_op_cap["weight"]["scheme"]:
                single_op_config["weight"]["scheme"] = single_op_cap["weight"]["scheme"][0]

            single_op_config["weight"]["granularity"] = (
                op_config.weight_granularity
                if op_config.weight_granularity in single_op_cap["weight"]["granularity"]
                else single_op_cap["weight"]["granularity"][0]
            )

            single_op_config["weight"]["algorithm"] = (
                op_config.weight_algorithm
                if op_config.weight_algorithm in single_op_cap["weight"]["algorithm"]
                else single_op_cap["weight"]["algorithm"][0]
            )

            op_wise_config.update({op_key_name: single_op_config})

        return op_wise_config

    def parse_to_tune_cfg(self):
        """The function that parses StaticQuantConfig to keras tuning config."""
        op_wise_config = self.update_opwise_config()
        tune_cfg = {"op": op_wise_config}

        return tune_cfg



# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The common module."""

from neural_compressor.common.utils import (
    level,
    level_name,
    logger,
    Logger,
    TuningLogger,
    log_process,
    set_random_seed,
    set_resume_from,
    set_workspace,
    set_tensorboard,
    dump_elapsed_time,
)
from neural_compressor.common.base_config import options


__all__ = [
    "options",
    "level",
    "level_name",
    "logger",
    "Logger",
    "TuningLogger",
    "log_process",
    "set_workspace",
    "set_random_seed",
    "set_resume_from",
    "set_tensorboard",
    "dump_elapsed_time",
]



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The constants utils for Tensorflow."""

SPR_BASE_VERSIONS = (
    "2.11.0202242",
    "2.11.0202250",
    "2.11.0202317",
    "2.11.0202323",
    "2.14.0202335",
    "2.14.dev202335",
    "2.15.0202341",
)

TENSORFLOW_DEFAULT_CONFIG = {
    "device": "cpu",
    "backend": "default",
    "approach": "post_training_static_quant",
    "random_seed": 1978,
    "format": "default",
    "use_bf16": True,
}

DEFAULT_SQ_ALPHA_ARGS = {
    "alpha_min": 0.0,
    "alpha_max": 1.0,
    "alpha_step": 0.1,
    "shared_criterion": "mean",
    "do_blockwise": False,
}

UNIFY_OP_TYPE_MAPPING = {
    "Conv2D": "conv2d",
    "Conv3D": "conv3d",
    "DepthwiseConv2dNative": "conv2d",
    "FusedBatchNormV3": "batchnorm",
    "FusedBatchNorm": "batchnorm",
    "_MklFusedInstanceNorm": "instancenorm",
    "MaxPool": "pooling",
    "MaxPool3D": "pooling",
    "AvgPool": "pooling",
    "ConcatV2": "concat",
    "MatMul": "matmul",
    "BatchMatMul": "matmul",
    "BatchMatMulV2": "matmul",
    "Pad": "pad",
    "Conv2DBackpropInput": "deconv2d",
    "Conv3DBackpropInputV2": "deconv3d",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initialize custom pool2d layers for Keras quantization."""

import json

import tensorflow as tf
from tensorflow.keras.layers import AveragePooling2D, MaxPooling2D

from neural_compressor.tensorflow.utils import version1_gte_version2


class QAvgPool2D(AveragePooling2D):
    """The custom quantized AveragePooling2D layer."""

    def __init__(
        self,
        name,
        pool_size=(2, 2),
        strides=None,
        padding="valid",
        data_format=None,
        act_min_value=None,
        act_max_value=None,
        weight_min_value=None,
        weight_max_value=None,
        granularity="per_tensor",
        quant_status="calib",
        quant_mode="SCALED",
        quant_T="s8",
        quant_round_mode="HALF_AWAY_FROM_ZERO",
        quant_narrow_range=False,
        quant_axis=None,
        **kwargs
    ):
        """Initialize custom quantized AveragePooling2D layer."""
        super(QAvgPool2D, self).__init__(
            name=name, pool_size=pool_size, strides=strides, padding=padding, data_format=data_format, **kwargs
        )
        T_map = {"s8": tf.qint8, "u8": tf.quint8}
        self.weight_min_value = weight_min_value
        self.weight_max_value = weight_max_value
        self.act_min_value = act_min_value
        self.act_max_value = act_max_value
        self.granularity = granularity
        self.quant_status = quant_status
        self.quant_mode = quant_mode
        self.quant_T = T_map[quant_T]
        self.quant_round_mode = quant_round_mode
        self.quant_narrow_range = quant_narrow_range
        self.quant_axis = quant_axis

    def __call__(self, inputs):
        """The __call__ function of custom quantized AveragePooling2D layer."""
        if self.quant_status == "calib" and not (
            version1_gte_version2(tf.__version__, "2.16.1") and isinstance(inputs, tf.keras.KerasTensor)
        ):
            if self.granularity == "per_tensor":
                self.act_min_value = tf.math.reduce_min(inputs)
                self.act_max_value = tf.math.reduce_max(inputs)
            else:
                self.act_min_value = tf.math.reduce_min(inputs, axis=1)
                self.act_max_value = tf.math.reduce_max(inputs, axis=1)
        elif self.quant_status == "quantize":
            inputs, _, _ = tf.quantization.quantize(
                inputs,
                self.act_min_value,
                self.act_max_value,
                self.quant_T,
                mode=self.quant_mode,
                round_mode=self.quant_round_mode,
                narrow_range=self.quant_narrow_range,
                axis=self.quant_axis,
            )
            inputs = tf.quantization.dequantize(
                inputs,
                self.act_min_value,
                self.act_max_value,
                mode=self.quant_mode,
                narrow_range=self.quant_narrow_range,
                axis=self.quant_axis,
            )

        return super(QAvgPool2D, self).__call__(inputs)

    @classmethod
    def from_config(cls, config):
        """Deserialize this class from a config dict."""
        return cls(**config)

    def get_config(self):
        """Serialize this class to a config dict."""
        config = super(QAvgPool2D, self).get_config()
        config.update(
            {
                "act_min_value": self.act_min_value,
                "act_max_value": self.act_max_value,
                "weight_min_value": self.weight_min_value,
                "weight_max_value": self.weight_max_value,
                "granularity": self.granularity,
                "quant_status": self.quant_status,
                "quant_mode": self.quant_mode,
                "quant_T": "s8" if self.quant_T == tf.qint8 else "u8",
                "quant_round_mode": self.quant_round_mode,
                "quant_narrow_range": self.quant_narrow_range,
                "quant_axis": self.quant_axis,
            }
        )

        return config


class QMaxPool2D(MaxPooling2D):
    """The custom quantized MaxPooling2D layer."""

    def __init__(
        self,
        name,
        pool_size=(2, 2),
        strides=None,
        padding="valid",
        data_format=None,
        act_min_value=None,
        act_max_value=None,
        weight_min_value=None,
        weight_max_value=None,
        granularity="per_tensor",
        quant_status="calib",
        quant_mode="SCALED",
        quant_T="s8",
        quant_round_mode="HALF_AWAY_FROM_ZERO",
        quant_narrow_range=False,
        quant_axis=None,
        **kwargs
    ):
        """Initialize custom quantized MaxPooling2D layer."""
        super(QMaxPool2D, self).__init__(
            name=name, pool_size=pool_size, strides=strides, padding=padding, data_format=data_format, **kwargs
        )
        T_map = {"s8": tf.qint8, "u8": tf.quint8}
        self.weight_min_value = weight_min_value
        self.weight_max_value = weight_max_value
        self.act_min_value = act_min_value
        self.act_max_value = act_max_value
        self.granularity = granularity
        self.quant_status = quant_status
        self.quant_mode = quant_mode
        self.quant_T = T_map[quant_T]
        self.quant_round_mode = quant_round_mode
        self.quant_narrow_range = quant_narrow_range
        self.quant_axis = quant_axis

    def __call__(self, inputs):
        """The __call__ function of custom quantized MaxPooling2D layer."""
        if self.quant_status == "calib" and not (
            version1_gte_version2(tf.__version__, "2.16.1") and isinstance(inputs, tf.keras.KerasTensor)
        ):
            if self.granularity == "per_tensor":
                self.act_min_value = tf.math.reduce_min(inputs)
                self.act_max_value = tf.math.reduce_max(inputs)
            else:
                self.act_min_value = tf.math.reduce_min(inputs, axis=1)
                self.act_max_value = tf.math.reduce_max(inputs, axis=1)
        elif self.quant_status == "quantize":
            assert self.act_min_value is not None, "Invalid activation min-max values, please check calibration process"
            inputs, _, _ = tf.quantization.quantize(
                inputs,
                self.act_min_value,
                self.act_max_value,
                self.quant_T,
                mode=self.quant_mode,
                round_mode=self.quant_round_mode,
                narrow_range=self.quant_narrow_range,
                axis=self.quant_axis,
            )
            inputs = tf.quantization.dequantize(
                inputs,
                self.act_min_value,
                self.act_max_value,
                mode=self.quant_mode,
                narrow_range=self.quant_narrow_range,
                axis=self.quant_axis,
            )

        return super(QMaxPool2D, self).__call__(inputs)

    @classmethod
    def from_config(cls, config):
        """Deserialize this class from a config dict."""
        return cls(**config)

    def get_config(self):
        """Serialize this class to a config dict."""
        config = super(QMaxPool2D, self).get_config()
        config.update(
            {
                "act_min_value": self.act_min_value,
                "act_max_value": self.act_max_value,
                "weight_min_value": self.weight_min_value,
                "weight_max_value": self.weight_max_value,
                "granularity": self.granularity,
                "quant_status": self.quant_status,
                "quant_mode": self.quant_mode,
                "quant_T": "s8" if self.quant_T == tf.qint8 else "u8",
                "quant_round_mode": self.quant_round_mode,
                "quant_narrow_range": self.quant_narrow_range,
                "quant_axis": self.quant_axis,
            }
        )

        return config


def initialize_int8_avgpool(fp32_layer, q_config):
    """Initialize int8 avgpool."""
    kwargs = fp32_layer.get_config()

    param_list = [
        "name",
        "pool_size",
        "strides",
        "padding",
        "data_format",
    ]
    for p in param_list:  # pragma: no cover
        if p in kwargs:
            del kwargs[p]

    q_layer = QAvgPool2D(
        name=fp32_layer.name,
        pool_size=fp32_layer.pool_size,
        strides=fp32_layer.strides,
        padding=fp32_layer.padding,
        data_format=fp32_layer.data_format,
        quant_T=q_config["T"],
        granularity=q_config["granularity"],
        **kwargs
    )

    return q_layer


def initialize_int8_maxpool(fp32_layer, q_config):
    """Initialize int8 maxpool."""
    kwargs = fp32_layer.get_config()

    param_list = [
        "name",
        "pool_size",
        "strides",
        "padding",
        "data_format",
    ]
    for p in param_list:  # pragma: no cover
        if p in kwargs:
            del kwargs[p]

    q_layer = QMaxPool2D(
        name=fp32_layer.name,
        pool_size=fp32_layer.pool_size,
        strides=fp32_layer.strides,
        padding=fp32_layer.padding,
        data_format=fp32_layer.data_format,
        quant_T=q_config["T"],
        granularity=q_config["granularity"],
        **kwargs
    )

    return q_layer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Post CSE Graph Rewriter."""

import hashlib

from tensorflow.core.framework import graph_pb2
from tensorflow.python.framework import tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class PostCseOptimizer(GraphRewriterBase):
    """Remove duplicated nodes like shared quantizev2 and const to decrease the output model size."""

    control_op_types = ("Switch", "Enter", "Merge", "NextIteration", "Exit")

    def _gen_node_hash(self, graph_info, node):
        """Generate nodes hash md5 data."""
        hash_str = node.op
        hash_str += str(len(node.input))
        for i in node.input:
            input_node = graph_info[Helper.node_name_from_input(i)].node
            if input_node.op == "Const":
                float_tensor = tensor_util.MakeNdarray(input_node.attr["value"].tensor)
                hash_str += str(float_tensor.flatten())
            else:
                hash_str += i

        attr_keys = sorted(node.attr)
        for i in attr_keys:
            hash_str += str(node.attr[i])

        return hashlib.md5(hash_str.encode("utf-8")).hexdigest()  # nosec

    @dump_elapsed_time("Pass PostCseOptimizer")
    def do_transformation(self):
        """Apply post CSE optimization."""
        GraphAnalyzer().graph = self.model
        graph_info = GraphAnalyzer().parse_graph()
        node_hash_info = {}
        loc_attr_node = []
        need_to_keep_const_node_name = []

        for _, v in graph_info.items():
            if "_class" in v.node.attr:
                loc_attr_node.append(v.node.attr["_class"].list.s[0].decode().split(":@")[-1])
        for node_name, i in graph_info.items():
            if node_name in loc_attr_node or i.node.op not in ("QuantizeV2", "Const"):
                continue

            hash_value = self._gen_node_hash(graph_info, i.node)

            if hash_value not in node_hash_info:
                node_hash_info[hash_value] = [node_name]

            if node_name not in node_hash_info[hash_value]:
                node_hash_info[hash_value].append(node_name)

        for _, v in node_hash_info.items():
            if len(v) == 1 or v[0] not in graph_info:
                continue

            node_type = graph_info[v[0]].node.op

            for j in v[1:]:
                if node_type == "Const" and j in graph_info:
                    output_op_types = [
                        graph_info[out_name].node.op in self.control_op_types for out_name in graph_info[j].outputs
                    ]
                    if any(output_op_types):
                        continue

                    for next_node in list(graph_info[j].outputs):
                        matched_index = 0
                        for index, origin_input in enumerate(graph_info[next_node].node.input):
                            if origin_input == j:
                                matched_index = index
                                break

                        graph_info[next_node].node.input[matched_index] = v[0]
                        if v[0] not in need_to_keep_const_node_name:
                            need_to_keep_const_node_name.append(v[0])
                        graph_info[v[0]].outputs.append(next_node)

                        if j in graph_info:
                            graph_info.pop(j)

                elif node_type == "QuantizeV2":  # pragma: no cover
                    next_node = graph_info[j].outputs[0]
                    quantize_v2_output_names = (j, j + ":1", j + ":2")

                    replace_index = [list(graph_info[next_node].node.input).index(i) for i in quantize_v2_output_names]

                    graph_info[next_node].node.input[replace_index[0]] = v[0]

                    graph_info[next_node].node.input[replace_index[1]] = v[0] + ":1"
                    graph_info[next_node].node.input[replace_index[2]] = v[0] + ":2"

                    graph_info[v[0]].outputs.append(next_node)

                    if graph_info[j].node.input[1] not in need_to_keep_const_node_name:
                        graph_info.pop(graph_info[j].node.input[1])

                    if graph_info[j].node.input[2] not in need_to_keep_const_node_name:
                        graph_info.pop(graph_info[j].node.input[2])

                    graph_info.pop(j)

                else:
                    self.logger.warning("Unknown Op type {}.".format(node_type))

        output_graph_def = graph_pb2.GraphDef()

        for _, node_info in graph_info.items():
            output_graph_def.node.extend([node_info.node])
        return output_graph_def



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The utility of common module."""

import collections
import enum
import importlib
import subprocess
import time
from typing import Dict

import cpuinfo
import psutil
from prettytable import PrettyTable

from neural_compressor.common.utils import Mode, TuningLogger, constants, logger

__all__ = [
    "set_workspace",
    "get_workspace",
    "set_random_seed",
    "set_resume_from",
    "set_tensorboard",
    "dump_elapsed_time",
    "log_process",
    "singleton",
    "LazyImport",
    "CpuInfo",
    "default_tuning_logger",
    "call_counter",
    "cpu_info",
    "ProcessorType",
    "detect_processor_type_based_on_hw",
    "Statistics",
]


def singleton(cls):
    """Singleton decorator."""
    instances = {}

    def _singleton(*args, **kw):
        """Create a singleton object."""
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton


class LazyImport(object):
    """Lazy import python module till use."""

    def __init__(self, module_name):
        """Init LazyImport object.

        Args:
           module_name (string): The name of module imported later
        """
        self.module_name = module_name
        self.module = None

    def __getattr__(self, name):
        """Get the attributes of the module by name."""
        try:
            self.module = importlib.import_module(self.module_name)
            mod = getattr(self.module, name)
        except:
            spec = importlib.util.find_spec(str(self.module_name + "." + name))
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        return mod

    def __call__(self, *args, **kwargs):
        """Call the function in that module."""
        function_name = self.module_name.split(".")[-1]
        module_name = self.module_name.split(f".{function_name}")[0]
        self.module = importlib.import_module(module_name)
        function = getattr(self.module, function_name)
        return function(*args, **kwargs)


@singleton
class CpuInfo(object):
    """Get CPU Info."""

    def __init__(self):
        """Get whether the cpu numerical format is bf16, the number of sockets, cores and cores per socket."""
        self._bf16 = False
        self._vnni = False
        info = cpuinfo.get_cpu_info()
        if "arch" in info and "X86" in info["arch"]:
            cpuid = cpuinfo.CPUID()
            max_extension_support = cpuid.get_max_extension_support()
            if max_extension_support >= 7:
                ecx = cpuid._run_asm(
                    b"\x31\xC9",  # xor ecx, ecx
                    b"\xB8\x07\x00\x00\x00" b"\x0f\xa2" b"\x89\xC8" b"\xC3",  # mov eax, 7  # cpuid  # mov ax, cx  # ret
                )
                self._vnni = bool(ecx & (1 << 11))
                eax = cpuid._run_asm(
                    b"\xB9\x01\x00\x00\x00",  # mov ecx, 1
                    b"\xB8\x07\x00\x00\x00" b"\x0f\xa2" b"\xC3",  # mov eax, 7  # cpuid  # ret
                )
                self._bf16 = bool(eax & (1 << 5))
        self._info = info
        self._brand_raw = info.get("brand_raw", "")
        # detect the below info when needed
        self._cores = None
        self._sockets = None
        self._cores_per_socket = None

    @property
    def brand_raw(self):
        """Get the brand name of the CPU."""
        return self._brand_raw

    @brand_raw.setter
    def brand_raw(self, brand_name):
        """Set the brand name of the CPU."""
        self._brand_raw = brand_name

    @staticmethod
    def _detect_cores():
        physical_cores = psutil.cpu_count(logical=False)
        return physical_cores

    @property
    def cores(self):
        """Get the number of cores in platform."""
        if self._cores is None:
            self._cores = self._detect_cores()
        return self._cores

    @cores.setter
    def cores(self, num_of_cores):
        """Set the number of cores in platform."""
        self._cores = num_of_cores

    @property
    def bf16(self):
        """Get whether it is bf16."""
        return self._bf16

    @property
    def vnni(self):
        """Get whether it is vnni."""
        return self._vnni

    @property
    def cores_per_socket(self) -> int:
        """Get the cores per socket."""
        if self._cores_per_socket is None:
            self._cores_per_socket = self.cores // self.sockets
        return self._cores_per_socket

    @property
    def sockets(self):
        """Get the number of sockets in platform."""
        if self._sockets is None:
            self._sockets = self._get_number_of_sockets()
        return self._sockets

    @sockets.setter
    def sockets(self, num_of_sockets):
        """Set the number of sockets in platform."""
        self._sockets = num_of_sockets

    def _get_number_of_sockets(self) -> int:
        if "arch" in self._info and "ARM" in self._info["arch"]:  # pragma: no cover
            return 1

        num_sockets = None
        cmd = "cat /proc/cpuinfo | grep 'physical id' | sort -u | wc -l"
        if psutil.WINDOWS:
            cmd = r'wmic cpu get DeviceID | C:\Windows\System32\find.exe /C "CPU"'
        elif psutil.MACOS:  # pragma: no cover
            cmd = "sysctl -n machdep.cpu.core_count"

        num_sockets = None
        try:
            with subprocess.Popen(
                args=cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=False,
            ) as proc:
                proc.wait()
                if proc.stdout:
                    for line in proc.stdout:
                        num_sockets = int(line.decode("utf-8", errors="ignore").strip())
        except Exception as e:
            logger.error("Failed to get number of sockets: %s" % e)
        if isinstance(num_sockets, int) and num_sockets >= 1:
            return num_sockets
        else:
            logger.warning("Failed to get number of sockets, return 1 as default.")
            return 1


cpu_info = CpuInfo()


def dump_elapsed_time(customized_msg=""):
    """Get the elapsed time for decorated functions.

    Args:
        customized_msg (string, optional): The parameter passed to decorator. Defaults to None.
    """

    def f(func):
        def fi(*args, **kwargs):
            start = time.time()
            res = func(*args, **kwargs)
            end = time.time()
            logger.info(
                "%s elapsed time: %s ms"
                % (customized_msg if customized_msg else func.__qualname__, round((end - start) * 1000, 2))
            )
            return res

        return fi

    return f


def set_random_seed(seed: int):
    """Set the random seed in config."""
    from neural_compressor.common import options

    options.random_seed = seed


def set_workspace(workspace: str):
    """Set the workspace in config."""
    from neural_compressor.common import options

    options.workspace = workspace


def get_workspace():
    """Get the workspace in config."""
    from neural_compressor.common import options

    return options.workspace


def set_resume_from(resume_from: str):
    """Set the resume_from in config."""
    from neural_compressor.common import options

    options.resume_from = resume_from


def set_tensorboard(tensorboard: bool):
    """Set the tensorboard in config."""
    from neural_compressor.common import options

    options.tensorboard = tensorboard


default_tuning_logger = TuningLogger()


def log_process(mode=Mode.QUANTIZE):
    """Decorator function that logs the stage of process.

    Args:
        mode (Mode): The mode of the process.

    Returns:
        The decorated function.
    """

    def log_process_wrapper(func):
        def inner_wrapper(*args, **kwargs):
            start_log = default_tuning_logger.execution_start
            end_log = default_tuning_logger.execution_end

            start_log(mode=mode, stacklevel=4)

            # Call the original function
            result = func(*args, **kwargs)

            end_log(mode=mode, stacklevel=4)

            return result

        return inner_wrapper

    return log_process_wrapper


# decorator for recording number of times a function is called
FUNC_CALL_COUNTS: Dict[str, int] = collections.defaultdict(int)


def call_counter(func):
    """A decorator that keeps track of the number of times a function is called.

    Args:
        func: The function to be decorated.

    Returns:
        The decorated function.
    """

    def wrapper(*args, **kwargs):
        FUNC_CALL_COUNTS[func.__name__] += 1
        return func(*args, **kwargs)

    return wrapper


class ProcessorType(enum.Enum):
    """The processor type."""

    Client = "Client"
    Server = "Server"


def detect_processor_type_based_on_hw():
    """Detects the processor type based on the hardware configuration.

    Returns:
        ProcessorType: The detected processor type (Server or Client).
    """
    # Detect the processor type based on below conditions:
    #   If there are more than one sockets, it is a server.
    #   If the brand name includes key word in `SERVER_PROCESSOR_BRAND_KEY_WORLD_LST`, it is a server.
    #   If the memory size is greater than 32GB, it is a server.
    log_mgs = "Processor type detected as {processor_type} due to {reason}."
    if cpu_info.sockets > 1:
        logger.info(log_mgs.format(processor_type=ProcessorType.Server.value, reason="there are more than one sockets"))
        return ProcessorType.Server
    elif any(brand in cpu_info.brand_raw for brand in constants.SERVER_PROCESSOR_BRAND_KEY_WORLD_LST):
        logger.info(
            log_mgs.format(processor_type=ProcessorType.Server.value, reason=f"the brand name is {cpu_info.brand_raw}.")
        )
        return ProcessorType.Server
    elif psutil.virtual_memory().total / (1024**3) > 32:
        logger.info(
            log_mgs.format(processor_type=ProcessorType.Server.value, reason="the memory size is greater than 32GB")
        )
        return ProcessorType.Server
    else:
        logger.info(
            "Processor type detected as %s, pass `processor_type='server'` to override it if needed.",
            ProcessorType.Client.value,
        )
        return ProcessorType.Client


class Statistics:
    """The statistics printer."""

    def __init__(self, data, header, field_names, output_handle=logger.info):
        """Init a Statistics object.

        Args:
            data: The statistics data
            header: The table header
            field_names: The field names
            output_handle: The output logging method
        """
        self.field_names = field_names
        self.header = header
        self.data = data
        self.output_handle = output_handle
        self.tb = PrettyTable(min_table_width=40)

    def print_stat(self):
        """Print the statistics."""
        valid_field_names = []
        for index, value in enumerate(self.field_names):
            if index < 2:
                valid_field_names.append(value)
                continue

            if any(i[index] for i in self.data):
                valid_field_names.append(value)
        self.tb.field_names = valid_field_names
        for i in self.data:
            tmp_data = []
            for index, value in enumerate(i):
                if self.field_names[index] in valid_field_names:
                    tmp_data.append(value)
            if any(tmp_data[1:]):
                self.tb.add_row(tmp_data)
        lines = self.tb.get_string().split("\n")
        self.output_handle("|" + self.header.center(len(lines[0]) - 2, "*") + "|")
        for i in lines:
            self.output_handle(i)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize ConcatV2 to int8 op."""

import os
import re

from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper

from ..quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithConcatV2(QuantizeNodeBase):
    """Quantize ConcatV2 to int8 op QuantizedConcatV2."""

    def __init__(self, **kwargs):
        """Initilizaiton."""
        super().__init__(**kwargs)
        self.sorted_patterns = sorted(self.patterns, key=lambda i: len(i), reverse=True)
        self.dtype = dtypes.quint8
        self.exclude_concat_nodes = []

    def _get_node_from_name(self, name):
        """Get node struct from node name."""
        if name.startswith("^"):
            name = name[1:]
        if re.search(r"\w+:\d+", name):
            node = self.node_name_mapping[name.rsplit(":", 1)[0]].node
        else:
            node = self.node_name_mapping[name].node
        return node

    def _get_first_input_from_name(self, name):
        """Get the first input of the node."""
        if len(name) == 0:
            return name
        node = self._get_node_from_name(name)
        if len(node.input) == 0:
            return ""
        return node.input[0]

    def _quantizable_concat(self, node):
        """Check if the ConcatV2 is quantizable."""
        deq_type = []
        is_quantizable = True
        if self.performance_only or os.getenv("TF_FORCE_CONCAT_OPTS") == "1":
            _, normal_inputs = self._get_node_input(node.name)
            original_inputs = normal_inputs[: node.attr["N"].i]

            # the input chain of concatv2 is QuantizedOp -> (req) -> q -> dq -> concat
            for each_input in original_inputs:
                dq_input = self._get_first_input_from_name(each_input)
                q_input = self._get_first_input_from_name(dq_input)
                pre_input = self._get_first_input_from_name(q_input)
                if pre_input == "":
                    continue
                req_input = self._get_node_from_name(pre_input)

                # the concatv2 with these Ops as inputs can't be reranged
                if req_input.op in ["_QuantizedFusedBatchNorm", "_QuantizedFusedInstanceNorm"]:
                    is_quantizable = False
                    break
                if (
                    req_input.op == "Requantize"
                    or req_input.op == "RequantizePerChannel"
                    or req_input.op.startswith("Quantized")
                ):
                    is_quantizable = True
                    if str(self.node_name_mapping[pre_input].node.attr["out_type"]) != "":
                        deq_type.append(self.node_name_mapping[pre_input].node.attr["out_type"].type)
                    else:
                        deq_type.append(self.node_name_mapping[pre_input].node.attr["T"].type)
        else:
            for input_node_name in node.input[: node.attr["N"].i]:
                node_name = helper.node_name_from_input(input_node_name)
                if self.node_name_mapping[node_name].node.op != "Dequantize":
                    self.exclude_concat_nodes.append(node.name)
                    return False

                deq_type.append(self.node_name_mapping[node_name].node.attr["T"].type)

        if len(set(deq_type)) != 1:
            is_quantizable = False
        else:
            if self.performance_only or os.getenv("TF_FORCE_CONCAT_OPTS") == "1":
                self.dtype = dtypes.DType(deq_type[0])

        if not is_quantizable:
            self.exclude_concat_nodes.append(node.name)

        return is_quantizable

    def _apply_concatv2_quantization(self, match_node_name):
        """Quantize ConcatV2."""
        # Dequantize + ConcatV2 + QuantizeV2
        skip_node_names = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]
        _, normal_inputs = self._get_node_input(matched_node.node.name)
        num_input = len(normal_inputs)
        shape_input_name = normal_inputs[num_input - 1]
        original_inputs = normal_inputs[0 : num_input - 1]

        input_names = []
        min_names = []
        max_names = []
        for each_input in original_inputs:
            _, q_each_input = self._get_node_input(each_input)
            input_names.append(q_each_input[0])
            min_names.append(q_each_input[1])
            max_names.append(q_each_input[2])
            skip_node_names.append(each_input)

        if self.dtype == dtypes.qint8:
            for each_name in input_names:
                input_node = self._get_node_from_name(each_name)
                if input_node.op == "QuantizeV2":
                    helper.set_attr_dtype(input_node, "T", self.dtype)
                pre_input = self._get_first_input_from_name(each_name)
                pre_input_node = self._get_node_from_name(pre_input)
                if pre_input_node.op == "Dequantize":
                    helper.set_attr_dtype(pre_input_node, "T", self.dtype)

        all_input_names = input_names
        all_input_names.append(shape_input_name)
        all_input_names.extend(min_names)
        all_input_names.extend(max_names)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_names:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))
                quantized_op_name = node.name + "_eightbit_quantized_concatv2"
                quantized_op_type = "Quantized" + node.op
                quantized_concat_node = helper.create_node(quantized_op_type, quantized_op_name, all_input_names)

                helper.set_attr_int(quantized_concat_node, "N", len(original_inputs))
                helper.set_attr_dtype(quantized_concat_node, "T", self.dtype)
                self.add_output_graph_node(quantized_concat_node)
                self._intel_cpu_add_dequantize_result_node(quantized_op_name, node.name, dtype=self.dtype)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self, do_transform=False):
        """Get longest fusion pattern."""
        self._get_op_list()
        matched_node_name = []

        for k, v in enumerate(self.op_list):
            if v in set(fusion[1] for fusion in self.sorted_patterns):
                cur_node = self.node_name_mapping[list(self.node_name_mapping.keys())[k]].node

                if cur_node.name != self.start_node_name:
                    continue

                # possible attributes that decide output data type
                output_attr_list = ["out_type", "T", "dtype", "Taxis", "Tindices", "Tparams"]
                if not do_transform:
                    _, normal_inputs = self._get_node_input(cur_node.name)
                    original_inputs = normal_inputs[: cur_node.attr["N"].i]
                    unsupported_input_type = False
                    for each_input in original_inputs:
                        each_input_name = helper.node_name_from_input(each_input)
                        input_dtype = None
                        for output_attr in output_attr_list:
                            input_dtype_attr = self.node_name_mapping[each_input_name].node.attr[output_attr]
                            if str(input_dtype_attr) != "":
                                input_dtype = dtypes.DType(input_dtype_attr.type)
                                break
                        if input_dtype != dtypes.bfloat16 and input_dtype != dtypes.float32:
                            unsupported_input_type = True
                            break
                    if unsupported_input_type:
                        continue

                for sub_rule in self.sorted_patterns:
                    if sub_rule[0] != "Dequantize" or sub_rule[-1] != "QuantizeV2":
                        continue
                    if v != sub_rule[1]:
                        continue

                if (self.performance_only or os.getenv("TF_FORCE_CONCAT_OPTS") == "1") and not do_transform:
                    matched_node_name.clear()
                    matched_node_name.append(cur_node.name)
                    return sub_rule, matched_node_name

                if self._quantizable_concat(cur_node):
                    if dtypes.as_dtype(cur_node.attr["T"].type) == dtypes.float32 and not re.search(
                        r"map(_\d+)?/while", cur_node.name
                    ):
                        matched_node_name.clear()
                        matched_node_name.append(sub_rule[0])
                        matched_node_name.append(cur_node.name)
                        matched_node_name.append(sub_rule[-1])
                        return sub_rule, matched_node_name
                else:
                    if self.performance_only or os.getenv("TF_FORCE_CONCAT_OPTS") == "1":
                        new_inputs = []
                        control_inputs, normal_inputs = self._get_node_input(cur_node.name)
                        original_inputs = normal_inputs[: cur_node.attr["N"].i]
                        for each_input in original_inputs:
                            each_node = self._get_node_from_name(each_input)
                            if each_node.op == "Dequantize":
                                q_input = self._get_first_input_from_name(each_input)
                                if q_input == "":
                                    continue
                                if self._get_node_from_name(q_input).op == "QuantizeV2":
                                    pre_input = self._get_first_input_from_name(q_input)
                                    new_inputs.append(pre_input)
                                elif self._get_node_from_name(q_input).op == "Requantize":
                                    new_inputs.append(each_input)
                            else:
                                new_inputs.append(each_input)
                        new_inputs.append(normal_inputs[-1])
                        cur_node.ClearField("input")
                        cur_node.input.extend(new_inputs + control_inputs)

        return None, None

    def apply_the_transform(self):
        """Apply the quantization of ConcatV2."""
        self._get_op_list()
        matched_rule, matched_node_name = self.get_longest_fuse(do_transform=True)
        if matched_node_name:
            fusion_name = "".join(matched_rule)
            if fusion_name == "DequantizeConcatV2QuantizeV2":
                self._apply_concatv2_quantization(matched_node_name)
            else:  # pragma: no cover
                self.logger.info("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph, self.exclude_concat_nodes

            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)
            return self.output_graph, self.exclude_concat_nodes

        if self.remove_redundant_quant_flag:
            self.input_graph = self.remove_redundant_quantization(self.input_graph)
        return self.input_graph, self.exclude_concat_nodes



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""Tensorflow Utils Helper functions."""

import os
from collections import OrderedDict, UserDict

import numpy as np
import tensorflow as tf
from google.protobuf import text_format
from tensorflow.core.framework import attr_value_pb2, graph_pb2, node_def_pb2, variable_pb2
from tensorflow.core.protobuf import config_pb2, meta_graph_pb2
from tensorflow.python.eager import context, wrap_function
from tensorflow.python.framework import convert_to_constants
from tensorflow.python.grappler import tf_optimizer
from tensorflow.python.platform import gfile
from tensorflow.python.saved_model import load, save, signature_constants, tag_constants
from tensorflow.python.training import saver
from tensorflow.python.util import nest

from neural_compressor.common import logger
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer, GraphRewriterHelper


def read_graph(in_graph, in_graph_is_binary=True):
    """Reads input graph file as GraphDef.

    :param in_graph: input graph file.
    :param in_graph_is_binary: whether input graph is binary, default True.
    :return: input graphDef.
    """
    assert gfile.Exists(in_graph), "Input graph pb file %s does not exist." % in_graph

    input_graph_def = graph_pb2.GraphDef()
    mode = "rb" if in_graph_is_binary else "r"
    with gfile.Open(in_graph, mode) as f:
        data = f.read()
        if in_graph_is_binary:
            input_graph_def.ParseFromString(data)
        else:
            text_format.Merge(data, input_graph_def)

    return input_graph_def


def write_graph(out_graph_def, out_graph_file):
    """Write output graphDef to file.

    :param out_graph_def: output graphDef.
    :param out_graph_file: path to output graph file.
    :return: None.
    """
    assert isinstance(out_graph_def, tf.compat.v1.GraphDef), "out_graph_def is not instance of TensorFlow GraphDef."

    assert out_graph_file and os.path.exists(
        os.path.dirname(out_graph_file)
    ), '"output_graph" directory does not exists.'

    f = gfile.GFile(out_graph_file, "wb")
    f.write(out_graph_def.SerializeToString())


def is_ckpt_format(model_path):
    """Check the model_path format is ckpt or not.

    Args:
        model_path (string): the model folder path

    Returns:
        string: return the ckpt prefix if the model_path contains ckpt format data else None.
    """
    file_list = [os.path.splitext(i)[-1] for i in os.listdir(model_path)]
    if file_list.count(".meta") == 1 and file_list.count(".index") == 1:
        return True
    return False


def _parse_ckpt_bn_input(graph_def):
    """Parse ckpt batch norm inputs to match correct moving mean and variance.

    Args:
        graph_def (graph_def): original graph_def
    Returns:
        graph_def: well linked graph_def
    """
    for node in graph_def.node:
        if node.op == "FusedBatchNorm":
            moving_mean_op_name = node.input[3]
            moving_var_op_name = node.input[4]
            moving_mean_op = _get_nodes_from_name(moving_mean_op_name, graph_def)[0]
            moving_var_op = _get_nodes_from_name(moving_var_op_name, graph_def)[0]

            if moving_mean_op.op == "Const":
                name_part = moving_mean_op_name.rsplit("/", 1)[0]
                real_moving_mean_op_name = name_part + "/moving_mean"
                if len(_get_nodes_from_name(real_moving_mean_op_name, graph_def)) > 0:
                    # replace the real moving mean op name
                    node.input[3] = real_moving_mean_op_name

            if moving_var_op.op == "Const":
                name_part = moving_var_op_name.rsplit("/", 1)[0]
                real_moving_var_op_name = name_part + "/moving_variance"
                if len(_get_nodes_from_name(real_moving_var_op_name, graph_def)) > 0:
                    # replace the real moving mean op name
                    node.input[4] = real_moving_var_op_name

    return graph_def


def _get_nodes_from_name(node_name, graph_def):
    """Get nodes from graph_def using node name.

    Args:
        graph_def (graph_def): graph_def
        node_name (str): node name

    Returns:
        node (NodeDef): graph node
    """
    return [node for node in graph_def.node if node.name == node_name]


def is_saved_model_format(model_path):
    """Check the model_path format is saved_model or not.

    Args:
        model_path (string): the model folder path

    Returns:
        bool: return True if the model_path contains saved_model format else False.
    """
    file_list = [os.path.splitext(i)[-1] for i in os.listdir(model_path)]
    # TF 2.11.0 added a new fingerprint.pb to the SavedModel directory.
    return bool(file_list.count(".pb") in [1, 2, 3] and ("variables") in os.listdir(model_path))


def get_tensor_by_name(graph, name, try_cnt=3):
    """Get the tensor by name.

    Considering the 'import' scope when model may be imported more then once,
    handle naming format like both name:0 and name.

    Args:
        graph (tf.compat.v1.GraphDef): the model to get name from
        name (string): tensor of tensor_name:0 or tensor_name without suffixes
        try_cnt: the times to add 'import/' to find  tensor

    Returns:
        tensor: tensor got by name.
    """
    if name.find(":") == -1:
        name = name + ":0"
    for _ in range(try_cnt):
        try:
            return graph.get_tensor_by_name(name)
        except BaseException:
            name = "import/" + name
    raise ValueError("can not find tensor by name")


def iterator_sess_run(sess, iter_op, feed_dict, output_tensor, iteration=-1, measurer=None):
    """Run the graph that have iterator integrated in the graph.

    Args:
        sess (tf.compat.v1.Session): the model sess to run the graph
        iter_op (Operator): the MakeIterator op
        feed_dict(dict): the feeds to initialize a new iterator
        output_tensor(list): the output tensors
        iteration(int): iterations to run, when -1 set, run to end of iterator

    Returns:
        preds: the results of the predictions
    """
    sess.run(iter_op, feed_dict)
    preds = []
    idx = 0
    while idx + 1 != iteration:
        try:
            if measurer:
                measurer.start()
                prediction = sess.run(output_tensor)
                measurer.end()
            else:
                prediction = sess.run(output_tensor)
            preds.append(prediction)
            idx += 1
        except tf.errors.OutOfRangeError:
            break

    preds = collate_tf_preds(preds)
    return preds


def collate_tf_preds(results):
    """Collate the prediction results."""
    batch = results[0]
    if isinstance(batch, list):
        results = zip(*results)
        collate_results = []
        for output in results:
            if isinstance(output[0], np.ndarray):
                collate_results.append(np.concatenate(output))
            elif np.isscalar(output[0]):
                collate_results.extend(output)
    elif isinstance(batch, np.ndarray):
        collate_results = np.concatenate(results)

    return collate_results


def get_input_output_node_names(graph_def):
    """Get the input node name and output node name of the graph_def."""
    g = GraphAnalyzer()
    g.graph = graph_def
    g.parse_graph()
    return g.get_graph_input_output()


def fix_ref_type_of_graph_def(graph_def):
    """Fix ref type of the graph_def."""
    # according to https://github.com/onnx/tensorflow-onnx/issues/77
    for node in graph_def.node:
        if node.op == "RefSwitch":
            node.op = "Switch"
            for index in range(len(node.input)):
                if "moving_" in node.input[index]:
                    node.input[index] = node.input[index] + "/read"
        elif node.op == "AssignSub":
            node.op = "Sub"
            if "use_locking" in node.attr:
                del node.attr["use_locking"]
        elif node.op == "AssignAdd":
            node.op = "Add"
            if "use_locking" in node.attr:
                del node.attr["use_locking"]
        elif node.op == "Assign":
            node.op = "Identity"
            if "use_locking" in node.attr:
                del node.attr["use_locking"]
            if "validate_shape" in node.attr:
                del node.attr["validate_shape"]
            if len(node.input) == 2:
                # input0: ref: Should be from a Variable node. May be uninitialized.
                # input1: value: The value to be assigned to the variable.
                node.input[0] = node.input[1]
                del node.input[1]
    return graph_def


def strip_unused_nodes(graph_def, input_node_names, output_node_names):
    """Strip unused nodes of the graph_def.

    The strip_unused_nodes pass is from tensorflow/python/tools/strip_unused_lib.py
    of official tensorflow r1.15 branch
    """
    cur_graph = GraphAnalyzer()
    cur_graph.graph = graph_def
    graph_info = cur_graph.parse_graph()
    type_attr = {"Sub": "T", "RealDiv": "T", "Identity": "T"}
    # this op should not be stripped for table initialization
    if "init_all_tables" in graph_info.keys():
        output_node_names.append("init_all_tables")
    not_found = {name for name in input_node_names}
    for node_name in list(graph_info.keys()):
        if node_name in not_found:
            not_found.remove(node_name)
            node = graph_info[node_name].node
            # skip the conversion to Placeholder that with type list
            if "component_types" in node.attr:
                continue
            original_output = graph_info[node_name].outputs
            placeholder_node = node_def_pb2.NodeDef()
            placeholder_node.op = "Placeholder"
            placeholder_node.name = node.name

            if "dtype" in node.attr:
                placeholder_node.attr["dtype"].CopyFrom(attr_value_pb2.AttrValue(type=node.attr["dtype"].type))
            elif node.op in type_attr.keys():
                placeholder_node.attr["dtype"].CopyFrom(
                    attr_value_pb2.AttrValue(type=node.attr[type_attr[node.op]].type)
                )
            else:
                raise KeyError("%s op's type attribute is not found," "you should add it to type_attr dict" % node.op)
            if "_output_shapes" in node.attr:
                placeholder_node.attr["_output_shapes"].CopyFrom(node.attr["_output_shapes"])
            if "shape" in node.attr:
                placeholder_node.attr["shape"].CopyFrom(node.attr["shape"])

            cur_graph.remove_node(node_name)

            cur_graph.replace_const_node(placeholder_node, [node_name], original_output)

    return tf.compat.v1.graph_util.extract_sub_graph(cur_graph.dump_graph(), output_node_names)


def strip_equivalent_nodes(graph_def, output_node_names):
    """Strip nodes with the same input and attr."""
    stripped_graph = GraphAnalyzer()
    stripped_graph.graph = graph_def
    stripped_graph_info = stripped_graph.parse_graph()

    def is_equivalent_input(input_tensor_list_1, input_tensor_list_2):
        if len(input_tensor_list_1) != len(input_tensor_list_2):
            return False
        const_num = 0
        for input_tensor_1, input_tensor_2 in zip(input_tensor_list_1, input_tensor_list_2):
            input_node_1 = stripped_graph_info[GraphRewriterHelper.node_name_from_input(input_tensor_1)].node
            input_node_2 = stripped_graph_info[GraphRewriterHelper.node_name_from_input(input_tensor_2)].node
            if input_node_1.op in ["Const", "HostConst"] and input_node_2.op in ["Const", "HostConst"]:
                if input_node_1.attr != input_node_2.attr:
                    return False
                const_num += 1
            elif input_tensor_1 != input_tensor_2:
                return False
        if const_num == len(input_tensor_list_1):
            return False
        return True

    nodes_to_remove = []
    replaced_nodes_type = {}
    stripped_graph_node_names = list(stripped_graph_info.keys())
    len_nodes = len(stripped_graph_node_names)
    for idx_1 in range(len_nodes - 1):
        node_name_1 = stripped_graph_node_names[idx_1]
        node_1 = stripped_graph_info[node_name_1].node
        if node_1.op in ["Const", "HostConst", "MatMul", "TensorArrayV3"] or node_name_1 in nodes_to_remove:
            continue
        for idx_2 in range(idx_1 + 1, len_nodes):
            node_name_2 = stripped_graph_node_names[idx_2]
            node_2 = stripped_graph_info[node_name_2].node
            if (
                node_1.op == node_2.op
                and node_name_1 != node_name_2
                and node_2 not in nodes_to_remove
                and node_1.input
                and is_equivalent_input(node_1.input, node_2.input)
                and node_1.attr == node_2.attr
            ):
                for ouput_node_name in stripped_graph_info[node_name_2].outputs:
                    output_node = stripped_graph_info[ouput_node_name].node
                    for idx_output_node_input, output_node_input_name in enumerate(output_node.input):
                        if GraphRewriterHelper.node_name_from_input(output_node_input_name) == node_name_2:
                            new_input = output_node_input_name.replace(node_name_2, node_name_1)
                            output_node.input[idx_output_node_input] = new_input
                            logger.debug(
                                "Replacing {} node '{}' with equivalent node '{}': "
                                "set {} node '{}'.input[{}] = '{}'".format(
                                    node_1.op,
                                    node_name_2,
                                    node_name_1,
                                    output_node.op,
                                    output_node.name,
                                    idx_output_node_input,
                                    new_input,
                                )
                            )
                            replaced_nodes_type[node_1.op] = replaced_nodes_type.get(node_1.op, 0) + 1
                            nodes_to_remove.append(node_name_2)
    for node_to_remove in nodes_to_remove:
        stripped_graph.remove_node(node_to_remove)
    return (
        tf.compat.v1.graph_util.extract_sub_graph(
            stripped_graph.dump_graph(), list(set(stripped_graph_node_names).intersection(output_node_names))
        ),
        replaced_nodes_type,
    )


# THIS API IS TO BE DEPRECATED!
def get_graph_def(model, outputs=[], auto_input_output=False):
    """Get the model's graph_def."""
    from neural_compressor.tensorflow.utils import BaseModel, Model

    if not isinstance(model, BaseModel):
        model = Model(model)
        model.output_tensor_names = outputs
    return model.graph_def


def get_model_input_shape(model):
    """Get the input shape of the input model."""
    for node in model.graph_def.node:
        if node.op == "Placeholder":
            _shape = list(tf.compat.v1.TensorShape(node.attr["shape"].shape))
            if tf.__version__ < "2.0.0":
                _shape = [item.value for item in _shape]
            if len(_shape) > 1 and isinstance(_shape[0], int):
                return _shape[0]
            elif isinstance(_shape, list) and hasattr(_shape[0], "value"):
                return _shape[0].value
    return 1


def generate_feed_dict(input_tensor, inputs):
    """Generate feed dict helper function."""
    if len(input_tensor) == 1:
        feed_dict = {}
        if isinstance(inputs, dict) or isinstance(inputs, OrderedDict) or isinstance(inputs, UserDict):
            for name in inputs:
                for tensor in input_tensor:
                    pos = tensor.name.rfind(":")
                    t_name = tensor.name if pos < 0 else tensor.name[:pos]
                    if name == t_name:
                        feed_dict[tensor] = inputs[name]
                        break
        else:
            feed_dict = {input_tensor[0]: inputs}  # get raw tensor using index [0]
    else:
        assert len(input_tensor) == len(inputs), "inputs len must equal with input_tensor"
        feed_dict = {}
        if isinstance(inputs, dict) or isinstance(inputs, OrderedDict) or isinstance(inputs, UserDict):
            for name in inputs:
                for tensor in input_tensor:
                    pos = tensor.name.rfind(":")
                    t_name = tensor.name if pos < 0 else tensor.name[:pos]
                    if name in [tensor.name, t_name]:
                        feed_dict[tensor] = inputs[name]
                        break
        else:
            # sometimes the input_tensor is not the same order with inputs
            # we should check and pair them
            def check_shape(tensor, data):
                # scalar or 1 dim default True
                if tensor.shape is None or len(tensor.shape.dims) == 1 or not hasattr(data, "shape"):
                    return True
                tensor_shape = tuple(tensor.shape)
                data_shape = tuple(data.shape)
                for tensor_dim, data_dim in zip(tensor_shape, data_shape):
                    if tensor_dim is not None and tensor_dim != data_dim:
                        return False
                return True

            disorder_tensors = []
            disorder_inputs = []
            for idx, sort_tensor in enumerate(input_tensor):
                sort_input = inputs[idx]
                if check_shape(sort_tensor, sort_input):
                    feed_dict.update({sort_tensor: sort_input})
                else:
                    disorder_tensors.append(sort_tensor)
                    disorder_inputs.append(sort_input)
            for i, dis_tensor in enumerate(disorder_tensors):
                for j, dis_input in enumerate(disorder_inputs):
                    if check_shape(dis_tensor, dis_input):
                        feed_dict.update({dis_tensor: dis_input})
                        break
    return feed_dict


def apply_inlining(func):
    """Apply an inlining optimization to the function's graph definition.

    Args:
        func: A concrete function get from saved_model.

    Returns:
        new_graph_def: The optimized graph in graph_def format.
    """
    graph_def = func.graph.as_graph_def()

    # In some cases, a secondary implementation of the function (e.g. for GPU) is
    # written to the "api_implements" attribute. (e.g. `tf.keras.layers.LSTM` in
    # TF2 produces a CuDNN-based RNN for GPU).
    # This function suppose to inline all functions calls, but "api_implements"
    # prevents this from happening. Removing the attribute solves the problem.
    # To learn more about "api_implements", see:
    #   tensorflow/core/grappler/optimizers/implementation_selector.h
    for function in graph_def.library.function:
        if "api_implements" in function.attr:
            del function.attr["api_implements"]

    meta_graph = saver.export_meta_graph(graph_def=graph_def, graph=func.graph)

    # Clear the initializer_name for the variables collections, since they are not
    # needed after saved to saved_model.
    for name in ["variables", "model_variables", "trainable_variables", "local_variables"]:
        raw_list = []
        for raw in meta_graph.collection_def["variables"].bytes_list.value:
            variable = variable_pb2.VariableDef()
            variable.ParseFromString(raw)
            variable.ClearField("initializer_name")
            raw_list.append(variable.SerializeToString())
        meta_graph.collection_def[name].bytes_list.value[:] = raw_list

    # Add a collection 'train_op' so that Grappler knows the outputs.
    fetch_collection = meta_graph_pb2.CollectionDef()
    for array in func.inputs + func.outputs:
        fetch_collection.node_list.value.append(array.name)
    meta_graph.collection_def["train_op"].CopyFrom(fetch_collection)

    # Initialize RewriterConfig with everything disabled except function inlining.
    config = config_pb2.ConfigProto()
    rewrite_options = config.graph_options.rewrite_options
    rewrite_options.min_graph_nodes = -1  # do not skip small graphs
    rewrite_options.optimizers.append("function")

    new_graph_def = tf_optimizer.OptimizeGraph(config, meta_graph)

    return new_graph_def


def construct_function_from_graph_def(func, graph_def, frozen_func=None):
    """Rebuild function from graph_def.

    Args:
        func: The original concrete function get from saved_model.
        graph_def: The optimized graph after applying inlining optimization.

    Returns:
        new_func: The reconstructed function.
    """
    if frozen_func is None:
        frozen_func = func

    # If a function is converted, then the TF context contains the original
    # function while the converted_graph_def contains the converted function.
    # Remove the original function from the TF context in this case.
    for f in graph_def.library.function:
        while context.context().has_function(f.signature.name):
            context.context().remove_function(f.signature.name)

    captures = {c[1].name.split(":")[0]: c[0] for c in frozen_func.graph.captures}
    new_func = wrap_function.function_from_graph_def(
        graph_def,
        [tensor.name for tensor in frozen_func.inputs],
        [tensor.name for tensor in frozen_func.outputs],
        captures,
    )
    new_func.graph.structured_outputs = nest.pack_sequence_as(
        func.graph.structured_outputs, new_func.graph.structured_outputs
    )
    # new_func._function_type = func.function_type  # pylint: disable=protected-access

    # Copy structured input signature from original function (used during
    # serialization)
    new_func.graph.structured_input_signature = func.structured_input_signature

    return new_func


def parse_saved_model(model, freeze=False, input_tensor_names=[], output_tensor_names=[]):
    """Parse a input saved_model.

    Args:
        model(string or AutoTrackable object): The input saved_model.

    Returns:
        graph_def: The graph_def parsed from saved_model.
        _saved_model: TF AutoTrackable object loaded from saved_model.
        func: The concrete function get from saved_model.
        frozen_func: The reconstructed function from inlining optimized graph.
    """
    config = tf.compat.v1.ConfigProto()
    config.use_per_session_threads = 1
    config.inter_op_parallelism_threads = 1

    if isinstance(model, str):
        _saved_model = load.load(model, [tag_constants.SERVING])
    else:
        _saved_model = model

    func = _saved_model.signatures[signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY]

    if freeze:
        frozen_func = convert_to_constants.convert_variables_to_constants_v2(func)
    else:
        inlined_graph_def = apply_inlining(func)
        frozen_func = construct_function_from_graph_def(func, inlined_graph_def)

    if len(input_tensor_names) == 0:
        # skip all inputs for ReadVariableOp
        input_tensor_names = [i.name.split(":")[0] for i in frozen_func.inputs if "unknown" not in i.name]
    if len(output_tensor_names) == 0:
        output_tensor_names = [i.name.split(":")[0] for i in frozen_func.outputs]

    frozen_graph_def = frozen_func.graph.as_graph_def()
    grappler_meta_graph_def = saver.export_meta_graph(graph_def=frozen_graph_def, graph=frozen_func.graph)

    # Add a collection 'train_op' so that Grappler knows the outputs.
    fetch_collection = meta_graph_pb2.CollectionDef()
    for array in frozen_func.inputs + frozen_func.outputs:
        fetch_collection.node_list.value.append(array.name)
        grappler_meta_graph_def.collection_def["train_op"].CopyFrom(fetch_collection)

    grappler_session_config = config_pb2.ConfigProto()
    rewrite_options = grappler_session_config.graph_options.rewrite_options
    rewrite_options.min_graph_nodes = -1
    graph_def = tf_optimizer.OptimizeGraph(grappler_session_config, grappler_meta_graph_def, graph_id=b"tf_graph")
    return graph_def, _saved_model, func, frozen_func, input_tensor_names, output_tensor_names


def reconstruct_saved_model(graph_def, func, frozen_func, trackable, path):
    """Reconstruct a saved_model.

    Args:
        graph_def: The input graph_def.
        func: The concrete function get from the original saved_model.
        frozen_func: The reconstructed function from inlining optimized graph.
        trackable: TF AutoTrackable object loaded from the original saved_model.
        path: The destination path to save the reconstructed saved_model.
    """
    converted_func = construct_function_from_graph_def(func, graph_def, frozen_func)
    signatures = {signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: converted_func}
    save.save(trackable, path, signatures, options=None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Model API for wrapping all TF model formats."""

import copy

from neural_compressor.common.utils import DEFAULT_WORKSPACE
from neural_compressor.tensorflow.utils.constants import TENSORFLOW_DEFAULT_CONFIG
from neural_compressor.tensorflow.utils.model_wrappers import BaseModel, KerasModel, TensorflowModel, get_tf_model_type
from neural_compressor.tensorflow.utils.utility import singleton


@singleton
class TensorflowGlobalConfig:
    """A global config class for setting framework specific information."""

    global_config = {
        "device": "cpu",
        "backend": "default",
        "approach": "post_training_static_quant",
        "random_seed": 1978,
        "workspace_path": DEFAULT_WORKSPACE,
        "format": "default",
        "use_bf16": True,
    }

    def reset_global_config(self):
        """Reset global config with default values."""
        self.global_config = copy.deepcopy(TENSORFLOW_DEFAULT_CONFIG)
        self.global_config["workspace_path"] = DEFAULT_WORKSPACE


TFConfig = TensorflowGlobalConfig()


class Model(object):  # pragma: no cover
    """A wrapper to construct a Neural Compressor TF Model."""

    def __new__(cls, root, **kwargs):
        """Create a new instance object of Model.

        Args:
            root (object): raw model format. For Tensorflow model, could be path to frozen pb file,
                path to ckpt or savedmodel folder, loaded estimator/graph_def/graph/keras model object.

        Returns:
            BaseModel: neural_compressor built-in model
        """
        from neural_compressor.tensorflow.utils import itex_installed

        if isinstance(root, BaseModel):
            return root

        if kwargs.get("approach", None) == "quant_aware_training":
            model_type = "keras_qat"
        elif "modelType" in kwargs:
            model_type = kwargs["modelType"]
        else:
            model_type = get_tf_model_type(root)

        if model_type == "keras" and not itex_installed():
            model_type = "saved_model"

        model = TensorflowModel(model_type, root, **kwargs)
        conf = kwargs.pop("conf", None)
        cls.set_tf_config(conf, model)

        return model

    @staticmethod
    def set_tf_config(conf, model):
        """Set tf global config with model information."""
        config = TFConfig.global_config
        framework = "keras" if isinstance(model, KerasModel) else "tensorflow"

        if conf and "device" in conf:
            config["device"] = conf["device"]
        if conf and "approach" in conf:
            config["approach"] = conf["approach"]
        if conf and "random_seed" in conf:
            config["random_seed"] = conf["random_seed"]
        if conf and "inputs" in conf:
            config["inputs"] = conf["inputs"]
        if conf and "outputs" in conf:
            config["outputs"] = conf["outputs"]

        if framework == "keras":
            config["backend"] = "itex"
            return

        from neural_compressor.tensorflow.utils import itex_installed

        if conf and "performance_only" in conf:
            config["performance_only"] = conf["performance_only"]
        if itex_installed():
            config["backend"] = "itex"
        if conf and "workspace_path" in conf:
            config["workspace_path"] = conf["workspace_path"]
        if conf and "recipes" in conf:
            config["recipes"] = conf["recipes"]

        for item in ["scale_propagation_max_pooling", "scale_propagation_concat"]:
            if "recipes" in config and item not in config["recipes"]:
                config["recipes"].update({item: True})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse QuantizedConv QuantizedDeConv with redundant Dequantize Graph Rewriter."""

from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

from ..graph_base import GraphRewriterBase


class FuseConvRedundantDequantizeTransformer(GraphRewriterBase):
    """Fuse _QuantizedConv/_QuantizedDeConv with the successor Dequantize Op."""

    fuse_patterns = [
        [
            "_FusedQuantizedConv3D",
            "_FusedQuantizedConv2D",
            "_FusedQuantizedDepthwiseConv2D",
            "_FusedQuantizedDeconv2D",
            "_FusedQuantizedDeconv3D",
        ],
        ["Dequantize"],
    ]

    fuse_sum_op_types_str = (
        str([b"BiasAdd", b"Sum", b"Requantize"]),
        str([b"BiasAdd", b"Sum", b"Relu", b"Requantize"]),
        str([b"BiasAdd", b"Sum", b"LeakyRelu", b"Requantize"]),
        str([b"BiasAdd", b"Relu", b"Sum", b"Requantize"]),
        str([b"BiasAdd", b"LeakyRelu", b"Sum", b"Requantize"]),
    )

    def __init__(self, model, device="cpu"):
        """Initialization."""
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.graph_info = self.graph_analyzer.parse_graph()

    def do_transformation(self):
        """Fuse the _QuantizedConv Op with the following Dequantize op.

        The output of _QuantizedConv or is fp32 or bf16.

        Returns:
            [graphdef]: the optimized graphdef object
        """
        dtype_map_dict = {
            dtypes.qint8.as_datatype_enum: dtypes.qint8,
            dtypes.quint8.as_datatype_enum: dtypes.quint8,
            dtypes.float32.as_datatype_enum: dtypes.float32,
            dtypes.qint32.as_datatype_enum: dtypes.qint32,
            dtypes.bfloat16.as_datatype_enum: dtypes.bfloat16,
        }
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.fuse_patterns)

        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            dequantize_node_name = i[1]
            dequantize_node = self.graph_info[dequantize_node_name].node

            if len(self.graph_info[quantized_node_name].outputs) > 3:
                continue

            # QuantizedConv doesn't support {"BiasAdd", "Sum", "Activation", "Dequantize"},
            # {"BiasAdd", "Activation", "Sum", "Dequantize"} and {"BiasAdd", "Sum", "Dequantize"}
            if str(quantized_node.attr["fused_ops"].list.s) in self.fuse_sum_op_types_str:
                continue

            new_node = node_def_pb2.NodeDef()
            new_node.op = quantized_node.op
            fused_ops = str(quantized_node.attr["fused_ops"].list.s).replace("Requantize", "Dequantize")
            new_node.name = quantized_node.name + "_dequantize"
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            if "Tinput" in quantized_node.attr:
                new_node.attr["Tinput"].CopyFrom(quantized_node.attr["Tinput"])
            if "Tfilter" in quantized_node.attr:
                new_node.attr["Tfilter"].CopyFrom(quantized_node.attr["Tfilter"])
            if "strides" in quantized_node.attr:
                new_node.attr["strides"].CopyFrom(quantized_node.attr["strides"])
            if "padding" in quantized_node.attr:
                new_node.attr["padding"].CopyFrom(quantized_node.attr["padding"])
            if "alpha" in quantized_node.attr:
                new_node.attr["alpha"].CopyFrom(quantized_node.attr["alpha"])
            if "Tbias" in quantized_node.attr:
                new_node.attr["Tbias"].CopyFrom(quantized_node.attr["Tbias"])
            if "data_format" in quantized_node.attr:
                new_node.attr["data_format"].CopyFrom(quantized_node.attr["data_format"])
            if "is_filter_const" in quantized_node.attr:
                new_node.attr["is_filter_const"].CopyFrom(quantized_node.attr["is_filter_const"])
            if "is_bias_const" in quantized_node.attr:
                new_node.attr["is_bias_const"].CopyFrom(quantized_node.attr["is_bias_const"])
            if "dilations" in quantized_node.attr:
                new_node.attr["dilations"].CopyFrom(quantized_node.attr["dilations"])
            if "explicit_paddings" in quantized_node.attr:
                new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr["explicit_paddings"])
            if "Tdevice_inputs" in quantized_node.attr:
                new_node.attr["Tdevice_inputs"].CopyFrom(quantized_node.attr["Tdevice_inputs"])
            if "Tdevice_outputs" in quantized_node.attr:
                new_node.attr["Tdevice_outputs"].CopyFrom(quantized_node.attr["Tdevice_outputs"])
            if "Thost_inputs" in quantized_node.attr:
                new_node.attr["Thost_inputs"].CopyFrom(quantized_node.attr["Thost_inputs"])
            Helper.set_attr_type_list(new_node, "Thost_outputs", [dequantize_node.attr["dtype"].type])
            new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=dequantize_node.attr["dtype"].type))
            Helper.set_attr_string_list(new_node, "fused_ops", eval(fused_ops))
            if "Tsummand" in quantized_node.attr:
                Helper.set_attr_dtype(new_node, "Tsummand", dtype_map_dict[dequantize_node.attr["dtype"].type])

            top_node_name = Helper.node_name_from_input(quantized_node.input[0])
            if self.graph_info[dequantize_node_name].outputs:
                self.graph_analyzer.replace_single_node(
                    new_node,
                    [top_node_name],
                    quantized_node_name,
                    self.graph_info[dequantize_node_name].outputs,
                    dequantize_node_name,
                )
                self.graph_analyzer.remove_node(dequantize_node_name)
            else:
                self.graph_analyzer.remove_node(dequantize_node_name)

                new_node.name = dequantize_node_name
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name, [], dequantize_node_name
                )

            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()
